import numpy as np
import pathlib
from collections import Counter
from typing import Tuple
import pyproj as pp
from vyperdatum.drivers.base import Driver

class NPZ(Driver):
    """
    Handle loading and parsing of a .npz (numpy arrays) file.

    Attributes
    ----------
    input_file: str
        Full path to the npz file.
    content: np.lib.npyio.NpzFile
        The object containing the numpy arrays stored in the npz file.

    Example
    --------
    >>> npz = NPZ('PATH_TO_NPZ_FILE')
    >>> x, y, z, u = npz.xyzu()
    >>> wkt = npz.wkt()
    >>> mmx, mmy, mmz, mmu = npz.minmax()
    """

    def __init__(self, input_file: str, invalid_error: bool = True) -> None:
        """
        Load a numpy .npz file (collection of numpy arrays).

        Parameters
        ----------
        input_file: str
            Full file path.
        invalid_error: bool, default True
            If True, throws an error when the input file has an unexpected format.

        Raises
        --------
        ValueError:
            If the input file is not recognized as npz file.

        Returns
        -----------
        None
        """
        super().__init__()
        self.input_file = input_file
        self.content = self.load()
        schema = ["wkt", "data", "minmax"]
        try:
            self.is_npz = Counter(self.content.files) == Counter(schema)
        except:
            self.is_npz = False
        if invalid_error and not self.is_npz:
            raise ValueError(("Expected the following keys in the .npz file: "
                              f"{schema}, but receieved {self.content.files}."))

    def load(self) -> np.lib.npyio.NpzFile:
        """
        Load a numpy .npz file (collection of numpy arrays).
        Return None when any exception take place or when the file is invalid. 

        Raises
        --------
        ValueError:
            If the file path is missing.
        FileNotFoundError:
            If the input file is not found.
        TypeError:
            If key names in the .npz file didn't match the expected names.

        Returns
        --------
        Optional[numpy.lib.npyio.NpzFile]
        """
        try:
            if not self.input_file:
                raise ValueError("Invalid or unspecified .npz file path.")
            input_file = pathlib.Path(self.input_file)
            if not input_file.is_file():
                raise FileNotFoundError(f"The npz file not found at: {input_file}")         
            bundle = np.load(input_file)
        except:
            return None
        return bundle


    def xyzu(self) -> Tuple[np.ndarray]:
        """
        Slice the `data` array to extract x, y, z, u arrays.

        Returns
        -------
        x: numpy.ndarray
            x coordinate numpy array.
        y: numpy.ndarray
            y coordinate numpy array.
        z: numpy.ndarray
            z coordinate numpy array.
        u: numpy.ndarray
            uncertainty numpy array.
        """
        if not self.content:
            raise ValueError("npz file not loaded.")
        x = self.content["data"][:, 0]
        y = self.content["data"][:, 1]
        z = self.content["data"][:, 2]
        u = self.content["data"][:, 3]
        return x, y, z, u

    def wkt(self) -> str:
        """
        Return wkt stored in the npz file.
        """
        if not self.content:
            raise ValueError("npz file not loaded.")
        return str(self.content["wkt"])

    def minmax(self) -> Tuple[np.ndarray]:
        """
        Extract the minmax values of the coordinated and uncertainty arrays.

        Returns
        -------
        x: numpy.ndarray
            x coordinate numpy array.
        y: numpy.ndarray
            y coordinate numpy array.
        z: numpy.ndarray
            z coordinate numpy array.
        u: numpy.ndarray
            uncertainty numpy array.
        """
        if not self.content:
            raise ValueError("npz file not loaded.")
        mmx = self.content["minmax"][:, 0]
        mmy = self.content["minmax"][:, 1]
        mmz = self.content["minmax"][:, 2]
        mmu = self.content["minmax"][:, 3]
        return mmx, mmy, mmz, mmu

    def transform(self, transformer_instance, vdatum_check: bool) -> bool:
        """
        Apply point transformation on npz data according to the `transformer_instance`.

        Parameters
        -----------
        transformer_instance: vyperdatum.transformer.Transform
            Instance of the transformer class.

        Returns
        -----------
        bool:
            True if successful, otherwise False.
        """
        x, y, z, u = self.xyzu()
        success, xx, yy, zz = transformer_instance.transform_points(x, y, z,
                                                                    vdatum_check=vdatum_check)
        target_wkt = transformer_instance.crs_to.to_wkt()
        data = np.zeros([len(x), 4], dtype=np.float64)
        data[:, 0], data[:, 1], data[:, 2], data[:, 3] = xx, yy, zz, u
        minmax = np.array([np.min(data, 0), np.max(data, 0)])
        np.savez(self.input_file, wkt=np.array(target_wkt), data=data, minmax=minmax)
        return success

    @property
    def is_valid(self):
        return self.is_npz



import pyproj as pp
from vyperdatum.utils import crs_utils as cu
import pytest


@pytest.mark.parametrize("crs_from, crs_to, vshift", [
    ("EPSG:6318", "EPSG:6319", False),
    ("EPSG:26919", "EPSG:6319", False),
    ("EPSG:26919", "EPSG:26919+EPSG:5703", True),
    ("EPSG:6318", "EPSG:9989", False),
    ("EPSG:6319", "EPSG:9989", False),
    ("EPSG:6319", "EPSG:4269", False),
    ("EPSG:6319", "EPSG:6318+NOAA:98", True),
    ("NOAA:1096", "NOAA:1098", True),
    ("EPSG:6318+EPSG:5703", "EPSG:9990+NOAA:101", True),
    ("EPSG:9755", "EPSG:6318", True),  # Is this a vertical shift? ellipsoids are not the same, therefore I suppose it should be considered as vertical shift
    ("EPSG:32618", "EPSG:9755", False),
    ])
def test_verical_shift(crs_from: str, crs_to: str, vshift: bool):
    ok = cu.vertical_shift(pp.CRS(crs_from), pp.CRS(crs_to)) == vshift
    assert ok, f"Vertical shift function for {crs_from} --> {crs_to} should be {not ok}"



import os
import pathlib
from osgeo import gdal, osr
import numpy as np
import pyproj as pp
import pytest
from vyperdatum.transformer import Transformer
from vyperdatum.drivers.npz import NPZ


def raster_wkt(raster_file: str):
    wkt = None
    if pathlib.Path(raster_file).suffix.lower() == ".npz":
        wkt = NPZ(raster_file).wkt()
    else:
        ds = gdal.Open(raster_file)
        srs = osr.SpatialReference(wkt=ds.GetProjection())
        wkt = srs.ExportToWkt()
        ds = None
    return wkt


# @pytest.mark.parametrize("input_file, bench_file, func", [
#     (r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\Modeling_BC25L26L_20230919.tiff",
#      r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\BlueTopo_BC25L26L_20230919.tiff",
#      "bluetopo")
#     ])
# def test_transform_bluetopo(input_file: str, bench_file: str, func: str):
#     xform_file = globals()[func](input_file)
#     gen_ds = gdal.Open(xform_file)
#     target_ds = gdal.Open(bench_file)
#     gen_band = np.nan_to_num(gen_ds.GetRasterBand(1).ReadAsArray())
#     target_band = np.nan_to_num(target_ds.GetRasterBand(1).ReadAsArray())
#     assert gen_ds.RasterCount == target_ds.RasterCount, "unexpected band counts"
#     assert pytest.approx(gen_band.min(), 0.001) == target_band.min(), f"inconsistent min band value (gen_min: {gen_band.min()} vs target_min: {target_band.min()})"
#     assert pytest.approx(gen_band.max(), 0.001) == target_band.max(), f"inconsistent max band value (gen_max: {gen_band.max()} vs target_max: {target_band.max()})"
#     gen_band.flags.writeable = False
#     target_band.flags.writeable = False
#     # assert hash(gen_band) == hash(target_band), f"hash check failed ({hash(gen_band)} vs {hash(target_band)})"
#     # assert gen_ds.GetRasterBand(1).Checksum() == target_ds.GetRasterBand(1).Checksum(), f"checksum failed ({gen_ds.GetRasterBand(1).Checksum()} vs {target_ds.GetRasterBand(1).Checksum()})"
#     # assert pp.CRS(raster_wkt(bench_file)).equals(pp.CRS(raster_wkt(xform_file))), "inconsistent crs."
#     gen_ds, target_ds = None, None
#     gen_band, target_band = None, None



@pytest.mark.parametrize("input_file, output_file, crs_from, crs_to", [
    (r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\00_test\Original\FL1701-TB-C_BLK-E-F_US4FL1ET_ellipsoidal_dem_b1.tif",
     r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\00_test\Manual\FL1701-TB-C_BLK-E-F_US4FL1ET_ellipsoidal_dem_b1.tif",
     "EPSG:6346",
     "EPSG:6346+NOAA:98",
     ),
    (r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\00_test\Original\ct1401_mllw_dem_1m.tif",
     r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\00_test\Manual\ct1401_mllw_dem_1m.tif",
     "EPSG:26919",
     "EPSG:26919+NOAA:98",
     ),
    (r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\00_test\Original\Modeling_BC25L26L_20230919.tiff",
     r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\00_test\Manual\Modeling_BC25L26L_20230919.tiff",
     "EPSG:26914+NOAA:98",
     "EPSG:26914+EPSG:5703",
     ),
    (r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\00_test\Original\MD1903-TB-C_US4MD1DD_ellipsoidal_dem.tif",
     r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\00_test\Manual\MD1903-TB-C_US4MD1DD_ellipsoidal_dem.tif",
     "EPSG:6347",
     "EPSG:6347+NOAA:98",
     )    
    ])
def test_transform_raster(input_file: str, output_file: str, crs_from: str, crs_to: str):
    tf = Transformer(crs_from=crs_from,
                     crs_to=crs_to
                     )
    success = tf.transform_raster(input_file=input_file,
                                  output_file=output_file,
                                  overview=False,
                                  )
    assert success, "Raster transformation unsuccessful."




# https://bag.readthedocs.io/en/master/howto-guide/index.html
# https://bag.readthedocs.io/en/master/fsd/FSD-Extension-VRGrid.html#xml-metadata-extensions-interpretation
# https://salishsea-meopar-tools.readthedocs.io/en/latest/bathymetry/ExploringBagFiles.html
# https://github.com/OpenNavigationSurface/BAG
# https://docs.h5py.org/en/stable/quick.html


import os
import time
import logging
import shutil
from typing import Optional
import concurrent.futures
import numpy as np
import h5py
from io import BytesIO
from lxml import etree
from tqdm.auto import tqdm
from osgeo import gdal
from pyproj.enums import WktVersion
from vyperdatum.utils.raster_utils import raster_metadata
from vyperdatum.enums import VRBAG as vrb_enum


logger = logging.getLogger("root_logger")
gdal.UseExceptions()


def is_vr(fname: str) -> bool:
    """
    Return True if fname points to a variable resolution BAG file.

    Parameters
    ----------
    fname: str
        Absolute path to the bag file.

    Returns
    ----------
    bool
    """
    try:
        bag = h5py.File(fname)
        vr = "varres_refinements" in list(bag["BAG_root"].keys())
        bag.close()
    except:
        vr = False
    return vr


def index_to_xy(i: int, j: int, geot: tuple, x_offset, y_offset):
    """
    Take indices of a cell from the base grid and return x, y.

    Parameters
    ----------
    i: int
        Cells's first index.
    j: int
        Cells's second index.
    geot: tuple
        Gdal GeoTransform tuple object.

    Returns
    ----------
    float, float
        easting, northing
    """
    res_x, res_y = geot[1], geot[5]
    x_min, y_max = geot[0], geot[3]
    x = j * res_x + x_min + x_offset
    y = i * res_y + y_max + y_offset
    return x, y


def base_grid_point_transform(fname: str,
                              tf,
                              nodata_value
                              ) -> np.ndarray:
    """
    Transform the low-resolution grid and return the transformed values.

    Parameters
    ----------
    fname: str
        Absolute path to the vrbag file.
    tf: vyperdatum.transformer.Transformer
        Instance of the transformer class.
    nodata_value: float
        No_Data_Value used in the vrbag elevation layer.

    Returns
    ----------
    np.ndarray
        The transformed vrbag elevation layer.
    """
    ds_gdal = gdal.Open(fname)
    geot = ds_gdal.GetGeoTransform()
    bag = h5py.File(fname)
    root = bag["BAG_root"]
    vr_elev = np.array(root["elevation"])
    vr_elev_shape = vr_elev.shape
    x, y, z = np.array([]), np.array([]), np.array([])
    for i in range(vr_elev_shape[0]):
        for j in range(vr_elev_shape[1]):
            _x, _y = index_to_xy(i, j, geot, x_offset=0, y_offset=0)
            x = np.append(x, [_x])
            y = np.append(y, [_y])
            z = np.append(z, vr_elev[i, j])

    _, _, _, zz = tf.transform_points(x, y, z)
    zz = np.where(z == nodata_value, z, zz)
    zz = zz.reshape(vr_elev_shape)

    ds_gdal = None
    bag.close()
    return zz


def update_vr_elevation(fname: str,
                        arr: np.ndarray) -> None:
    """
    Update the `elevation` layer in the vrbag file with the `arr` values.

    Parameters
    ----------
    fname : str
        Absolute path to the vrbag file.
    arr: np.ndarray
        numpy array representing the elevation values to be updated.

    Returns
    ----------
    None
    """

    bag = h5py.File(fname, "r+")
    root = bag.require_group("/BAG_root")
    del root["elevation"]
    root.create_dataset("elevation",
                        maxshape=(None, None),
                        data=np.array(arr, dtype=np.float32),
                        fillvalue=vrb_enum.NDV_REF.value,
                        compression="gzip",
                        compression_opts=9
                        )
    arr = np.where(arr == vrb_enum.NDV_REF.value, np.nan, arr)
    root["elevation"].attrs.create("Maximum Elevation Value", np.nanmax(arr), dtype=np.float32)
    root["elevation"].attrs.create("Minimum Elevation Value", np.nanmin(arr), dtype=np.float32)
    bag.close()
    return


def get_subgrid_points(fname: str, i: int, j: int) -> tuple[list[int], list[float],
                                                            list[float], list[float]]:
    """
    Return the starting index and coordinates of the points of subgrid i, j.

    Parameters
    ----------
    fname: str
        Absolute path to the vrbag file.
    i: int
        First index of the subgrid.
    j: int
        Second index of the subgrid.

    Returns
    ----------
    list[int], list[float], list[float], list[float]
        Starting index of subgrid.
        x, y, z coordinates of the subgrid points.
    """
    try:
        er = "00"

        ds_gdal = gdal.Open(fname)
        er = "01"
        geot = ds_gdal.GetGeoTransform()
        er = "02"
        bag = h5py.File(fname)
        er = "03"
        root = bag["BAG_root"]
        er = "04"
        vr_meta = root["varres_metadata"]
        er = "05"
        vr_ref = root["varres_refinements"][0]

        er = "06"
        start = vr_meta[i, j][0]
        dim_x, dim_y = vr_meta[i, j][1], vr_meta[i, j][2]
        res_x, res_y = vr_meta[i, j][3], vr_meta[i, j][4]
        sw_corner_x, sw_corner_y = vr_meta[i, j][5], vr_meta[i, j][6]
        cell_x, cell_y = index_to_xy(i, j, geot, x_offset=sw_corner_x, y_offset=sw_corner_y)

        er = "07"

        x = np.array([cell_x + (i - i // dim_x) * res_x for i in range(dim_x*dim_y)])
        y = np.array([cell_y + (i // dim_x) * res_y for i in range(dim_x*dim_y)])
        z = np.array([vr[0] for vr in vr_ref[start:start+(dim_x*dim_y)]])
        
        er = "08"
        
        ds_gdal = None
        bag.close()
    except Exception as e:
        print("SSSSSSSSSSSSSSSSSSSSSS")
        print(er)
        logger.exception(f"Unexpected exception in get_subgrid_points for subgrid {i}, {j}: {e}")
        f = open("error.txt", "a")
        f.write(f"Unexpected exception in get_subgrid_points for subgrid {i}, {j}: {e}\n")
        f.close()
        start, x, y, z = None, None, None, None
    return start, x, y, z


def single_subgrid_point_transform(fname: str,
                                   i: int,
                                   j: int,
                                   tf,
                                   nodata_value
                                   ) -> tuple[Optional[int], Optional[np.ndarray]]:
    """
    Apply point transformation of subgrid points.

    Parameters
    ----------
    fname: str
        Absolute path to the vrbag file.
    i: int
        The first index of the subgrid.
    j: int
        The second index of the subgrid.
    tf: vyperdatum.transformer.Transformer
        Instance of the transformer class.
    nodata_value: float
        No_Data_Value used for the generated GeoTiff.

    Returns
    ----------
    tuple[Optional[int], np.ndarray]
        The starting index of the subgrid in the varres_refinements layer.
        The transformed subgrid depth values in form of a 1-d array.
    """
    try:
        start, x, y, z = get_subgrid_points(fname, i, j)
        _, _, _, zz = tf.transform_points(x, y, z)
        zz = np.where(z == nodata_value, z, zz)
    except Exception as e:
        logger.exception(f"Unexpected exception in single_subgrid_point_transform for subgrid {i}, {j}: {e}")
        f = open("error.txt", "a")
        f.write(f"Unexpected exception in single_subgrid_point_transform for subgrid {i}, {j}: {e}\n")
        f.close()
        start, zz = None, None

    return start, zz


def subgrid_point_transform(fname: str,
                            tf,
                            ) -> tuple[list[int], list[float]]:
    """
    Identify the subgrids within the vrbag and return the starting
    index and the depth values within each subgrid.

    Parameters
    ----------
    fname: str
        Absolute path to the vrbag file.
    tf: vyperdatum.transformer.Transformer
        Instance of the transformer class.

    Returns
    ----------
    list[int], list[float]
        Starting index of each subgrid.
        Transformed subgrid depth values.
    """
    bag = h5py.File(fname)
    root = bag["BAG_root"]
    vr_meta = root["varres_metadata"]
    start_indices, transformed_refs, ii, jj = [], [], [], []
    for i in tqdm(range(vr_meta.shape[0]), desc="Search for refined subgrids"):
        for j in range(vr_meta.shape[1]):
            start = vr_meta[i, j][0]
            if start == vrb_enum.NO_REF_INDEX.value:
                continue
            ii.append(i)
            jj.append(j)
    bag.close()

    ii = ii[:3]
    jj = jj[:3]
    with concurrent.futures.ThreadPoolExecutor() as executor:
        futureObjs = executor.map(single_subgrid_point_transform,
                                  [fname] * len(ii),
                                  ii, jj,
                                  [tf] * len(ii),
                                  [vrb_enum.NDV_REF.value] * len(ii)
                                  )
        for i, fo in enumerate(tqdm(futureObjs, total=len(ii))):
            if fo[0] is not None:
                start_indices.append(fo[0])
                transformed_refs.append(fo[1])


    # for i in tqdm(range(len(ii))):    
    #     s, z = single_subgrid_point_transform(fname, ii[i], jj[i], tf, vrb_enum.NDV_REF.value)
    #     if s is not None:
    #         start_indices.append(s)
    #         transformed_refs.append(z)
    # # print(start_indices, transformed_refs)    
    return start_indices, transformed_refs


def single_subgrid_rsater_transform(fname: str,
                                    rasters_dir: str,
                                    i: int,
                                    j: int,
                                    tf,
                                    nodata_value
                                    ) -> tuple[Optional[int], Optional[np.ndarray]]:
    """
    Extract a subgrid from from the vrbag file, convert to GeoTiff, and apply transformation.

    Parameters
    ----------
    fname: str
        Absolute path to the vrbag file.
    rasters_dir: str
        Absolute path to the directory where the output TIFF files will be stored.
    i: int
        First index of the subgrid.
    j: int
        Second index of the subgrid.
    tf: vyperdatum.transformer.Transformer
        Instance of the transformer class.
    nodata_value: float
        No_Data_Value used for the generated GeoTiff.

    Returns
    ----------
        The starting index of the subgrid in the varres_refinements layer.
        The transformed subgrid in form of a 1-d array.
    """
    try:
        bm = raster_metadata(fname)
        geot = bm["geo_transform"]
        wkt_str = bm["wkt"]

        bag = h5py.File(fname)
        root = bag["BAG_root"]
        vr_meta = root["varres_metadata"]
        vr_ref = root["varres_refinements"][0]
        start = vr_meta[i, j][0]

        dim_x, dim_y = vr_meta[i, j][1], vr_meta[i, j][2]
        res_x, res_y = vr_meta[i, j][3], vr_meta[i, j][4]
        sw_corner_x, sw_corner_y = vr_meta[i, j][5], vr_meta[i, j][6]

        sub_x_min = geot[0] + j * geot[1] + sw_corner_x
        sub_y_min = geot[3] + i * geot[5] + sw_corner_y
        sub_extent = [sub_x_min, sub_y_min, sub_x_min + geot[1], sub_y_min + abs(geot[5])]
        sub_geot = (sub_extent[0], res_x, 0, sub_extent[3], 0, -res_y)
        sub_grid = vr_ref[start:start+(dim_x*dim_y)]["depth"].reshape((dim_y, dim_x))

        # Save subgrid as Geotiff
        driver = gdal.GetDriverByName("GTiff")
        sub_raster_fname = f"{rasters_dir}{i}_{j}.tiff"
        out_ds = driver.Create(sub_raster_fname, sub_grid.shape[1],
                               sub_grid.shape[0], 1, gdal.GDT_Float32)
        out_ds.SetProjection(wkt_str)
        out_ds.SetGeoTransform(sub_geot)
        band = out_ds.GetRasterBand(1)
        band.WriteArray(sub_grid)
        band.SetNoDataValue(nodata_value)
        band.FlushCache()
        band.ComputeStatistics(False)
        out_ds = None

        #####################
        transformed_sub_fname = f"{rasters_dir}t_{i}_{j}.tiff"
        tf.transform_raster(input_file=sub_raster_fname, output_file=transformed_sub_fname,
                            pre_post_checks=False, vdatum_check=False)
        ds = gdal.Open(transformed_sub_fname)
        transformed_refs = ds.GetRasterBand(1).ReadAsArray().flatten()
        ds = None
        bag.close()
        gdal.Unlink(sub_raster_fname)
        gdal.Unlink(transformed_sub_fname)

        if len(transformed_refs) != dim_y * dim_x:
            msg = f"Transferred length for sub array {i}, {j} was expected to be {dim_y * dim_x} but is {len(transformed_refs)}\n"
            logger.error(msg)
            f = open("error.txt", "a")
            f.write(msg)
            f.close()

    except Exception as e:
        logger.exception(f"Unexpected exception in single_subgrid_rsater_transform: {e}")

        f = open("error.txt", "a")
        f.write(f"Unexpected exception in single_subgrid_rsater_transform for subgrid {i}, {j}: {e}\n")
        f.close()

        start, transformed_refs = None, None
    return start, transformed_refs


def subgrid_raster_transform(fname: str,
                             rasters_dir: str,
                             tf,
                             ) -> tuple[list[int], list[float]]:
    """
    Identify the subgrids within the vrbag and return the starting
    index and the depth values within each subgrid.

    Parameters
    ----------
    fname : str
        Absolute path to the vrbag file.
    rasters_dir : str
        Absolute path to the directory where the output TIFF files will be stored.
    tf: vyperdatum.transformer.Transformer
        Instance of the transformer class.

    Returns
    ----------
    list[int], list[float]
        Starting index of each subgrid.
        Transformed subgrid depth values.
    """

    if rasters_dir.split("/")[0].lower() != "vsimem":
        if os.path.isdir(rasters_dir):
            shutil.rmtree(rasters_dir)
        os.makedirs(rasters_dir)
    bag = h5py.File(fname)
    root = bag["BAG_root"]
    vr_meta = root["varres_metadata"]
    start_indices, transformed_refs, ii, jj = [], [], [], []
    for i in tqdm(range(vr_meta.shape[0]), desc="Making subgrid rasters"):
        for j in range(vr_meta.shape[1]):
            start = vr_meta[i, j][0]
            if start == vrb_enum.NO_REF_INDEX.value:
                continue
            ii.append(i)
            jj.append(j)
    bag.close()
    with concurrent.futures.ProcessPoolExecutor() as executor:
        futureObjs = executor.map(single_subgrid_rsater_transform,
                                  [fname] * len(ii),
                                  [rasters_dir] * len(ii),
                                  ii, jj,
                                  [tf] * len(ii),
                                  [vrb_enum.NDV_REF.value] * len(ii)
                                  )
        for i, fo in enumerate(futureObjs):
            if fo[0] is not None:
                start_indices.append(fo[0])
                transformed_refs.append(fo[1])
    return start_indices, transformed_refs


def corner_points(fname: str) -> tuple[float, float, float, float]:
    """
    Return the corner points indicated in the xml metadata.

    Parameters
    ----------
    fname : str
        Absolute path to the vrbag file.

    Returns
    ----------
    str
        corner points: x1, y1, x2, y2
    """
    bag = h5py.File(fname)
    meta = bag["BAG_root/metadata"]
    buffer = BytesIO(meta[()])
    tree = etree.parse(buffer)
    root = tree.getroot()
    gml = ".//{" + root.nsmap['gml'] + "}"
    p1, p2 = root.find(f"{gml}coordinates").text.split(" ")
    x1, y1 = p1.split(",")
    x2, y2 = p2.split(",")
    return float(x1), float(y1), float(x2), float(y2) 


def change_corner_points_and_wkt(fname: str,
                                 new_points: str,
                                 wkt_h: str,
                                 wkt_v: str
                                 ) -> None:
    """
    Update the xml metadata's corner points, horizontal, and vertical WKTs with new values.

    Parameters
    ----------
    fname : str
        Absolute path to the vrbag file.
    new_points: str
        The new corners points in string format x1,y1 x2,y2. Will be ignored when None.
    wkt_h: str
        WKT string for the horizontal component of the CRS. Will be ignored when None.
    wkt_v: str
        WKT string for the vertical component of the CRS. Will be ignored when None.

    Returns
    ----------
    None
    """
    bag = h5py.File(fname)
    meta = bag["BAG_root/metadata"]
    buffer = BytesIO(meta[()])
    tree = etree.parse(buffer)
    root = tree.getroot()
    gml = ".//{" + root.nsmap['gml'] + "}"
    gco = ".//{" + root.nsmap['gco'] + "}"
    root.find(f"{gml}coordinates").text = new_points
    root.findall(f"{gco}CharacterString")[6].text = wkt_h
    root.findall(f"{gco}CharacterString")[7].text = "WKT"
    root.findall(f"{gco}CharacterString")[8].text = wkt_v
    root.findall(f"{gco}CharacterString")[9].text = "WKT"
    # tree.write(xml_fname)
    # xml = etree.tostring(root, pretty_print=True).decode("ascii")
    xmet = etree.tostring(root).decode()
    bag.close()
    bag = h5py.File(fname, mode="r+")
    root = bag.require_group("/BAG_root")
    del bag["/BAG_root/metadata"]
    metadata = np.array(list(xmet), dtype="S1")
    root.create_dataset("metadata",
                        maxshape=(None,),
                        data=metadata,
                        compression="gzip",
                        compression_opts=9
                        )
    bag.close()
    return


def update_vr_refinements(fname: str,
                          index: list[int],
                          arr: list[np.ndarray],
                          tf) -> None:
    """
    Update the `varres_refinements` layer in the vrbag file with the
    `arr` values starting form `index` location in the `varres_refinements`.

    Parameters
    ----------
    fname : str
        Absolute path to the vrbag file.
    index: list[int]
        A list of starting index where the refinements get updated.
    arr: list[np.ndarray]
        List of numpy array representing the refinements values to be updated.
    tf: vyperdatum.transformer.Transformer
        Instance of the transformer class.

    Returns
    ----------
    None
    """
    print("\nUpdating varres_refinements layer...\n")
    bag = h5py.File(fname, "r+")
    root = bag.require_group("/BAG_root")
    vr_ref = root["varres_refinements"]
    vr_ref_type = [("depth", np.float32), ("depth_uncrt", np.float32)]
    vr_ref = np.array(vr_ref, dtype=vr_ref_type)
    for i, index_start in enumerate(index):
        vr_ref[0][index_start:len(arr[i])+index_start]["depth"] = arr[i]
    del root["varres_refinements"]
    root.create_dataset("varres_refinements",
                        maxshape=(1, None),
                        data=vr_ref,
                        # fillvalue=np.array([(vrb_enum.NDV_REF.value, vrb_enum.NDV_REF.value)], dtype=vr_ref_type),
                        compression="gzip",
                        compression_opts=9
                        )
    darr = np.array(vr_ref[0][:]["depth"])
    uarr = np.array(vr_ref[0][:]["depth_uncrt"])
    darr = np.where(darr == vrb_enum.NDV_REF.value, np.nan, darr)
    uarr = np.where(uarr == vrb_enum.NDV_REF.value, np.nan, uarr)
    root["varres_refinements"].attrs.create("max_depth", np.nanmax(darr), dtype=np.float32)
    root["varres_refinements"].attrs.create("max_uncrt", np.nanmax(uarr), dtype=np.float32)
    root["varres_refinements"].attrs.create("min_depth", np.nanmin(darr), dtype=np.float32)
    root["varres_refinements"].attrs.create("min_uncrt", np.nanmin(uarr), dtype=np.float32)
    bag.close()
    # update the base grid and its attributes
    zt = base_grid_point_transform(fname=fname, tf=tf, nodata_value=vrb_enum.NDV_REF.value)
    update_vr_elevation(fname=fname, arr=zt)
    # update xml
    x1, y1, x2, y2 = corner_points(fname=fname)
    # _, y1, x1, _ = tf.transform_points([y1], [x1], [0], always_xy=False, allow_ballpark=False)
    # _, y2, x2, _ = tf.transform_points([y2], [x2], [0], always_xy=False, allow_ballpark=False)

    _, x1, y1, _ = tf.transform_points([y1], [x1], [0], always_xy=False, allow_ballpark=False)
    _, x2, y2, _ = tf.transform_points([y2], [x2], [0], always_xy=False, allow_ballpark=False)


    # using old WKT1_GDAL for compatibility with older GDAL/QGIS
    if tf.crs_to.is_compound:
        wkt_h = tf.crs_to.sub_crs_list[0].to_wkt(version=WktVersion.WKT1_GDAL)
        wkt_v = tf.crs_to.sub_crs_list[1].to_wkt(version=WktVersion.WKT1_GDAL)
    else:
        wkt_h = tf.crs_to.to_wkt(version=WktVersion.WKT1_GDAL)
        wkt_v = ""
    wkt_h = wkt_h if wkt_h else ""
    wkt_v = wkt_v if wkt_v else ""
    change_corner_points_and_wkt(fname=fname,
                                 new_points=f"{x1[0]},{y1[0]} {x2[0]},{y2[0]}",
                                 wkt_h=wkt_h,
                                 wkt_v=wkt_v
                                 )
    return


def transform(fname: str,
              tf,
              vdatum_check: bool = True,
              point_transformation: bool = True,
              **kwargs
              ) -> bool:
    """
    Transform vrbag according to the `tf` Transformer object.
    When `point_transformation` is True, point transformation is applied, otherwise
    vrbag is split into rasters and raster transformation is applied. When raster transformation
    is chosen, a new keyword argument `rasters_dir` must be passed that specifies where the subgrid
    rasters are stored.

    Parameters
    ----------
    fname : str
        Absolute path to the vrbag file.
    index: list[int]
        A list of starting index where the refinements get updated.
    arr: list[np.ndarray]
        List of numpy array representing the refinements values to be updated.
    tf: vyperdatum.transformer.Transformer
        Instance of the transformer class.
    vdatum_check: bool, default=True
        If True, a random sample of the transformed data are compared with transformation
        outcomes produced by Vdatum REST API.
    rasters_dir: str
        Absolute path to the directory where the output TIFF files will be stored.
        When raster transformation is chosen, this parameter must be passed
        which specifies where the subgrid rasters are stored. It can point to a local
        dir (e.g. `rasters_dir = "./sub_grids/"`) or GDAL virtual file system
        (e.g. `rasters_dir = "/vsimem/sub_grids/"`).

    Raises
    ----------
    TypeError
        If the passed BAG file is not a valid variable resolution bag file.
    KeyError
        When raster transformation is chosen (`point_transformation = False`), and `rasters_dir`
        parameter is not passed to the function.

    Returns
    ----------
    bool:
        True if successful, otherwise False.
    """
    try:
        success = False
        if not is_vr(fname=fname):
            msg = (f"The following file is not a valid variable resolution bag file: {fname}")
            logger.exception(msg)
            raise TypeError(msg)
        if not point_transformation and "rasters_dir" not in kwargs.keys():
            msg = ("For raster transformation approach, you must pass `rasters_dir` parameter"
                   " to the `transform_vr` function.")
            logger.exception(msg)
            raise KeyError(msg)
        tic = time.time()
        if point_transformation:
            index, zt = subgrid_point_transform(fname, tf=tf)
        else:
            index, zt = subgrid_raster_transform(fname=fname,
                                                 rasters_dir=kwargs["rasters_dir"],
                                                 tf=tf
                                                 )
        update_vr_refinements(fname=fname, index=index, arr=zt, tf=tf)
        logger.info(f"VRBAG transformation processing time: {time.time() - tic:.2f}", )
        success = True
    except Exception as e:
        logger.exception(f"Unexpected exception in transform_vr: {e}")
        return False
    return success


def wkt(fname: str) -> str:
    """
    Return wkt retrieved by GDAL.
    """
    return raster_metadata(fname)["wkt"]



import os
from vyperdatum.transformer import Transformer
from vyperdatum.utils.raster_utils import raster_metadata, update_raster_wkt
from vyperdatum.utils.vdatum_rest_utils import vdatum_cross_validate
import pyproj as pp
from osgeo import gdal


def get_tiff_files(parent_dir: str, extention: str) -> list:
    tiff_files = []
    for (dirpath, dirnames, filenames) in os.walk(parent_dir):
        for filename in filenames:
            if filename.endswith(extention):
                tiff_files.append(os.sep.join([dirpath, filename]))
    return tiff_files


def remove_second_band(src_path: str, dst_path: str) -> str:
    """
    Remove band 2 from a raster and write the output to a new folder,
    preserving spatial reference, geotransform, and metadata.

    Notes
    -----
    - Overviews/pyramids are not copied; rebuild with `gdaladdo` if needed.
    - If the source has fewer than 2 bands, this raises a ValueError.
    """
    gdal.UseExceptions()

    ds = gdal.Open(src_path, gdal.GA_ReadOnly)
    if ds is None:
        raise RuntimeError(f"Failed to open {src_path}")

    band_count = ds.RasterCount
    if band_count < 2:
        raise ValueError(f"Input has only {band_count} band(s); nothing to drop.")

    # Build band list: keep band 1 and bands 3..N
    band_list = [1] + list(range(3, band_count + 1))

    os.makedirs(os.path.split(dst_path)[0], exist_ok=True)
    driver_name = "GTiff"

    # Translate with the band selection
    translate_opts = gdal.TranslateOptions(
        bandList=band_list,
        format=driver_name,
        creationOptions=["TILED=YES", "COMPRESS=LZW"]
    )

    out_ds = gdal.Translate(dst_path, ds, options=translate_opts)
    if out_ds is None:
        raise RuntimeError("gdal.Translate failed to create the output dataset.")

    # (Optional) sanity pass to ensure band-level attributes are carried over.
    # gdal.Translate typically copies these, but we enforce them just in case.
    # Map: src bands 1,3,4,... -> out bands 1,2,3,...
    mapping = {src_idx: out_idx for out_idx, src_idx in enumerate(band_list, start=1)}
    for src_idx, out_idx in mapping.items():
        sb = ds.GetRasterBand(src_idx)
        ob = out_ds.GetRasterBand(out_idx)

        # NoData
        ndv = sb.GetNoDataValue()
        if ndv is not None:
            ob.SetNoDataValue(ndv)

        # Scale/offset/units/description
        scale = sb.GetScale()
        if scale is not None:
            ob.SetScale(scale)
        offset = sb.GetOffset()
        if offset is not None:
            ob.SetOffset(offset)
        unit = sb.GetUnitType()
        if unit:
            ob.SetUnitType(unit)
        desc = sb.GetDescription()
        if desc:
            ob.SetDescription(desc)

        # Color table & interpretation
        ct = sb.GetColorTable()
        if ct is not None:
            ob.SetColorTable(ct)
        interp = sb.GetColorInterpretation()
        if interp is not None:
            ob.SetColorInterpretation(interp)

        # Band metadata
        ob.SetMetadata(sb.GetMetadata())

    out_ds.FlushCache()
    out_ds = None
    ds = None
    return dst_path




if __name__ == "__main__":
    parent_dir = r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\PBB\Original\FL2205-TB-C"
    parent_dir = r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\PBB\Original\FL1812-TB-N"
    parent_dir = r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\PBB\Original\2020_ngs_topobathyDEM_michael_J1219754"
    parent_dir = r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\PBB\Original\3_band_dem"
    files = get_tiff_files(parent_dir, extention=".tif")
    
    # files = [
    #     # r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\PBB\Original\3_band_dem\2018_525000e_2780000n_tpu.tif",
    #     r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\PBB\Original\3_band_dem\2018_525000e_2785000n_tpu.tif",
    #     r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\PBB\Original\3_band_dem\2018_530000e_2785000n_tpu.tif",
    #     # r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\PBB\Original\3_band_dem\2018_535000e_2775000n_tpu.tif",
    #     # r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\PBB\Original\3_band_dem\2018_550000e_2785000n_tpu.tif",
    #     ]
    crs_from = "EPSG:6346"
    crs_to = "EPSG:6346+NOAA:98"
    # crs_to = "EPSG:26917"

    # steps = [{'crs_from': 'EPSG:6346', 'crs_to': 'EPSG:6318', 'v_shift': False},
    #          {'crs_from': 'EPSG:6318+EPSG:5703', 'crs_to': 'EPSG:6319', 'v_shift': True},
    #          {'crs_from': 'EPSG:6319', 'crs_to': 'EPSG:6318+NOAA:98', 'v_shift': True},
    #          {'crs_from': 'EPSG:6318', 'crs_to': 'EPSG:6346', 'v_shift': False}
    #          ]
    for i, input_file in enumerate(files[:]):
        print(f"{i+1}/{len(files)}: {input_file}")
        tf = Transformer(crs_from=crs_from,
                         crs_to=crs_to,
                        #  steps=steps
                         )
        
        prc_file = input_file
        # prc_file = input_file.replace("Original", "Processed")
        # remove_second_band(src_path=input_file, dst_path=prc_file)
        output_file = input_file.replace("Original", "Manual")
        tf.transform_raster(input_file=prc_file,
                            output_file=output_file,
                            overview=False,
                            pre_post_checks=True,
                            vdatum_check=False
                            )                 
        print(f'\n{"*"*50} {i+1}/{len(files)} Completed {"*"*50}\n')



import os
import sys
sys.path.insert(0, os.path.abspath(".."))
sys.path.insert(0, os.path.abspath("../vyperdatum"))

project = "Vyperdatum"
copyright = "2024, Mohammad Ashkezari"
author = "Mohammad Ashkezari"
release = "0.3.4"

extensions = ["sphinx.ext.todo", "sphinx.ext.viewcode", "sphinx.ext.autodoc", "notfound.extension"]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]



from vyperdatum.transformer import Transformer


if __name__ == "__main__":
    in_fname = r"C:\Users\mohammad.ashkezari\Desktop\laz\ma2021_cent_east_Job1082403.laz"
    out_fname = r"C:\Users\mohammad.ashkezari\Desktop\laz\transformed_ma2021_cent_east_Job1082403.laz"

    crs_from = "EPSG:6348+EPSG:5703"
    crs_to = "EPSG:6348+NOAA:5320"
    steps = ["EPSG:6348+EPSG:5703", "EPSG:6318+EPSG:5703", "EPSG:6318+NOAA:5320", "EPSG:6348+NOAA:5320"]
    tf = Transformer(crs_from=crs_from, crs_to=crs_to, steps=steps)
    tf.transform(input_file=in_fname, output_file=out_fname)



import os
from vyperdatum.transformer import Transformer


fname = r"C:\Users\mohammad.ashkezari\Desktop\original_vrbag\W00656_MB_VR_MLLW_5of5.bag"
crs_from = "EPSG:32617+EPSG:5866"
crs_to = "EPSG:26917+EPSG:5866"
steps = ["EPSG:32617+EPSG:5866", "EPSG:9755", "EPSG:6318", "EPSG:26917+EPSG:5866"]
tf = Transformer(crs_from=crs_from,
                 crs_to=crs_to,
                #  steps=steps
                 )
d, f = os.path.split(fname)
output_file = os.path.join(d, f"transformed_{f}")
tf.transform(input_file=fname, output_file=output_file)



import pytest
import os
import pyproj as pp
from vyperdatum.transformer import Transformer
from vyperdatum.drivers.npz import NPZ
from vyperdatum.drivers.laz import LAZ
from vyperdatum.drivers.gparq import GeoParquet



# def test_6349_to_6319():
#     """
#     Verify the a point on NAD83_2011_NAVD88 (EPSG:6349) is transformed
#     correctly to NAD83_2011_3D (EPSG:6319).

#     If network is off then a noop is used for the transform and NAVD88 to
#     NAD83_2011_3D returns 0.0 leading to the test failure.
#     """

#     coords_lat_lon = (39, -76.5, 0)
#     x, y, z = Transformer(crs_from=6349, crs_to=6319).transform_points(*coords_lat_lon)
#     assert pytest.approx(x, abs=.01) == coords_lat_lon[0], "x coordinate should remain unchanged."
#     assert pytest.approx(y, abs=.01) == coords_lat_lon[1], "y coordinate should remain unchanged."
#     assert pytest.approx(z, abs=.01) == -33.29, "incorrect z coordinate transformation."


# @pytest.mark.parametrize("fname", [
#     r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\point\npz\2018_NCMP_MA_19TCF2495_BareEarth_1mGrid_transformed.bruty.bruty.npz",
#     r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\point\npz\W00405_MB_1m_MLLW_1of4_transformed.bruty.npz"
# ])
# def test_npz(fname):
#     npz = NPZ(fname)
#     x, y, z, u = npz.xyzu()
#     wkt = npz.wkt()
#     mmx, mmy, mmz, mmu = npz.minmax()
#     assert x.shape == y.shape == z.shape == u.shape, f"inconsistent data array dimensions in npz file: {fname}"  # noqa: E501
#     assert mmx.shape == mmy.shape == mmz.shape == mmu.shape, f"inconsistent minmax array dimensions in npz file: {fname}"  # noqa: E501
#     assert isinstance(wkt, str), f"unexpected wkt type in npz file: {fname}"


def test_transform_gparq():
    fname = r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\point\geoparquet\South_Locust_Point_&_Masonville_&_Ferry_Bar_&_Fairfield_2023_Elevation_MLLW_3ftx3ft.parquet"
    d, f = os.path.split(fname)
    output_file = os.path.join(d, f"transformed_{f}")
    crs_from = "EPSG:6347+NOAA:98"
    crs_to = "EPSG:6347+EPSG:5703"
    tf = Transformer(crs_from=crs_from, crs_to=crs_to)
    tf.transform(input_file=fname, output_file=output_file)
    gp = GeoParquet(output_file)
    auth_code = "+".join([":".join(pp.CRS(s).to_authority()) for s in pp.CRS(gp.wkt()).sub_crs_list])
    assert auth_code == crs_to, f"Incorrect WKT in the transformed geoparquet file: {output_file}"


# def test_transform_npz():
#     fname = r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\point\npz\W00405_MB_1m_MLLW_1of4_transformed.bruty.npz"
#     d, f = os.path.split(fname)
#     output_file = os.path.join(d, f"transformed_{f}")
#     crs_from = "EPSG:26919"
#     crs_to = "EPSG:26919+NOAA:5434"
#     steps = ["EPSG:26919", "EPSG:6319", "EPSG:6318+NOAA:5434", "EPSG:26919+NOAA:5434"]
#     tf = Transformer(crs_from=crs_from, crs_to=crs_to, steps=steps)
#     tf.transform(input_file=fname, output_file=output_file)
#     npz = NPZ(output_file)
#     auth_code = "+".join([":".join(pp.CRS(s).to_authority()) for s in pp.CRS(npz.wkt()).sub_crs_list])
#     assert auth_code == crs_to, f"Incorrect WKT in the transformed npz file: {output_file}"


# def test_transform_laz():
#     fname = r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\point\laz\ma2021_cent_east_Job1082403.laz"
#     d, f = os.path.split(fname)
#     output_file = os.path.join(d, f"transformed_{f}")
#     crs_from = "EPSG:6348+EPSG:5703"
#     crs_to = "EPSG:6348+NOAA:5320"
#     steps = ["EPSG:6348+EPSG:5703", "EPSG:6318+EPSG:5703", "EPSG:6318+NOAA:5320", "EPSG:6348+NOAA:5320"]
#     tf = Transformer(crs_from=crs_from, crs_to=crs_to, steps=steps)
#     tf.transform(input_file=fname, output_file=output_file)
#     lz = LAZ(output_file)
#     auth_code = "+".join([":".join(pp.CRS(s).to_authority()) for s in pp.CRS(lz.wkt()).sub_crs_list])
#     assert auth_code == crs_to, f"Incorrect WKT in the transformed laz file: {output_file}"


# def test_transform_vrbag():
#     fname = r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\point\vrbag\W00656_MB_VR_MLLW_5of5.bag"
#     d, f = os.path.split(fname)
#     output_file = os.path.join(d, f"transformed_{f}")
#     crs_from = "EPSG:32617+EPSG:5866"
#     crs_to = "EPSG:26917+EPSG:5866"
#     steps = ["EPSG:32617+EPSG:5866", "EPSG:9755", "EPSG:6318", "EPSG:26917+EPSG:5866"]
#     tf = Transformer(crs_from=crs_from, crs_to=crs_to, steps=steps)
#     tf.transform(input_file=fname, output_file=output_file)

#     auth_code = "+".join([":".join(pp.CRS(s).to_authority()) for s in pp.CRS(wkt).sub_crs_list])
#     assert auth_code == crs_to, f"Incorrect WKT in the transformed vrbag file: {output_file}"


import os
import rasterio
from rasterio.transform import Affine
from rasterio.crs import CRS
import numpy as np
import glob
import pathlib
from osgeo import gdal

gdal.UseExceptions()

def create_geotiff_from_tif_tfw(tif_path, tfw_path, output_path, epsg_code, nodata):
    """
    Creates a georeferenced GeoTIFF from a TIFF and its world file (.tfw).

    Parameters:
        tif_path (str): Path to the .tif image file.
        tfw_path (str): Path to the .tfw world file.
        output_path (str): Path to write the output GeoTIFF.
        epsg_code (int): EPSG code representing the CRS.
        nodata (float or int): NoData value to set in the output file.

    Returns:
        output_path (str): Path to the generated GeoTIFF.
    """
    # Read the .tfw world file
    with open(tfw_path, 'r') as f:
        A = float(f.readline())
        D = float(f.readline())
        B = float(f.readline())
        E = float(f.readline())
        C = float(f.readline())
        F = float(f.readline())

    transform = Affine(A, B, C, D, E, F)
    crs = CRS.from_epsg(epsg_code)

    with rasterio.open(tif_path) as src:
        data = src.read()
        profile = src.profile.copy()
        profile.update({
            'transform': transform,
            'crs': crs,
        })

        # if nodata is not None:
        #     profile.update({'nodata': nodata})

        with rasterio.open(output_path, 'w', **profile) as dst:
            dst.write(data)

    update_stats(output_path)
    return output_path


def update_stats(input_file):
    """
    Update statistics for the raster file.

    Parameters:
        input_file (str): Path to the raster file.
    """
    ds = gdal.Open(input_file, gdal.GA_Update)
    for i in range(1, ds.RasterCount + 1):
        ds.GetRasterBand(i).ComputeStatistics(False)
    ds = None  # flush
    return

files = glob.glob(r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\ALK\Original\**\*.tif", recursive=True)
files = glob.glob(r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\ALK\Manual2\**\*.tif", recursive=True)

for i, input_file in enumerate(files[:]):
    print(f"{i+1}/{len(files)}: {input_file}")
    update_stats(input_file)
    # output_file = input_file.replace("Original", "Processed")
    # tfw_file = input_file.replace(".tif", ".tfw")
    # pathlib.Path(os.path.split(output_file)[0]).mkdir(parents=True, exist_ok=True)
    # create_geotiff_from_tif_tfw(input_file, tfw_file, output_file, 6337, -999999)


import logging
import subprocess
from colorama import Fore, Style
from typing import Optional, Tuple
import pyproj as pp
from pyproj.transformer import TransformerGroup


logger = logging.getLogger("root_logger")


def crs_components(crs: pp.CRS, raise_no_auth: bool = True) -> Tuple[str, Optional[str]]:
    """
    Return CRS horizontal and vertical components string representation
    in form of code:authority. If the input CRS is horizontal, the vertical
    component will be None.

    Raises
    -------
    ValueError:
        If either code or authority of the crs (or its sub_crs) can not be determined.

    Returns
    --------
    Tuple[str, Optional[str]]:
        Horizontal and vertical components of crs in form of code:authority
    """
    if isinstance(crs, str):
        crs = pp.CRS(crs)
    h, v = None, None
    if crs.is_compound:
        try:
            sub_h = pp.CRS(crs.sub_crs_list[0])
            # if sub_h.is_bound:
            #     sub_h = sub_h.source_crs
            h = ":".join(sub_h.to_authority())
        except:
            if sub_h.is_bound:
                h = "BOUNDCRS"
            else:
                h = "UnknownAuthorityCode"
        try:
            sub_v = pp.CRS(crs.sub_crs_list[1])
            # if sub_v.is_bound:
            #     sub_v = sub_v.source_crs
            v = ":".join(sub_v.to_authority())
        except:
            if sub_v.is_bound:
                v = "BOUNDCRS"
            else:
                v = "UnknownAuthorityCode"
    else:
        # if crs.is_bound:
        #     crs = crs.source_crs
        ac = crs.to_authority(min_confidence=100)
        if not ac and raise_no_auth:
            raise ValueError(f"Unable to produce authority name and code for this crs:\n{crs}")
        h = ":".join(crs.to_authority())
    return h, v


def auth_code(crs: pp.CRS, raise_no_auth: bool = True) -> Optional[str]:
    """
    Return CRS string representation in form of code:authority

    Raises
    -------
    ValueError:
        If either code or authority of the crs (or its sub_crs) can not be determined.

    Returns
    --------
    str:
        crs string in form of code:authority
    """
    if isinstance(crs, str):
        crs = pp.CRS(crs)
    ## might fail for pyproj.CRS("EPSG:4269").geodetic_crs.to_3d().to_authority(min_confidence=100)
    ac = crs.to_authority(min_confidence=100)
    if not ac and crs.is_compound:
        try:
            h = ":".join(pp.CRS(crs.sub_crs_list[0]).to_authority())
        except:
            h = "UnknownAuthorityCode"
        try:
            v = ":".join(pp.CRS(crs.sub_crs_list[1]).to_authority())
        except:
            v = "UnknownAuthorityCode"
        return f"{h}+{v}"
    if not ac and raise_no_auth:
        raise ValueError(f"Unable to produce authority name and code for this crs:\n{crs}")
    return ":".join(ac)


def vertical_shift(crs_from: pp.CRS, crs_to: pp.CRS) -> bool:
    """
    Return True if transformation from `crs_from` to `crs_to` results
    in vertical shifts.

    Parameters
    ----------
    crs_from: pyproj.crs.CRS
        Projection of input data.
    crs_to: pyproj.crs.CRS
        Projection of output data.

    Raises
    -------
    TypeError:
        If either one of `crs_from` or `crs_to` is not of type pyproj.CRS.

    Returns
    -------
    bool
    """
    if not isinstance(crs_from, pp.CRS) or not isinstance(crs_to, pp.CRS):
        raise TypeError("Both `crs_from` and `crs_to` must be of type pyproj.CRS.")
    if crs_from.equals(crs_to):
        return False
    vertical = False
    # if (not crs_from.is_projected and not crs_to.is_projected
    #     and len(crs_from.axis_info) + len(crs_to.axis_info) == 5):  # 2D+3D
    #     vertical = True
    # if len(crs_from.axis_info) + len(crs_to.axis_info) == 6:  # 3D + 3D
    #     s_v = crs_from.sub_crs_list[1] if crs_from.is_compound else None
    #     t_v = crs_to.sub_crs_list[1] if crs_to.is_compound else None
    #     if s_v is None and t_v is None and crs_from.datum.ellipsoid != crs_to.datum.ellipsoid:
    #         vertical = True
    #     elif s_v != t_v:
    #         vertical = True

    if crs_from.is_compound or crs_to.is_compound:
        s_v = crs_from.sub_crs_list[1] if crs_from.is_compound else None
        t_v = crs_to.sub_crs_list[1] if crs_to.is_compound else None
        if s_v != t_v:
            vertical = True
    elif (crs_from.datum.ellipsoid and  # WGS84-based CRS datum's ellipsoid is None
          crs_to.datum.ellipsoid and
          crs_from.datum.ellipsoid != crs_to.datum.ellipsoid):
        vertical = True
    return vertical


def crs_epoch(crs: pp.CRS) -> Optional[str]:
    """
    Return the input CRS reference epoch, if the input CRS is Dynamic.
    otherwise return `None`.

    Parameters
    ----------
    crs: pyproj.crs.CRS
        pyproj CRS instance.

    Raises
    -------
    TypeError:
        If `crs` is not of type pyproj.CRS.
    """
    if not isinstance(crs, pp.CRS):
        raise TypeError("`crs` must be of type pyproj.CRS.")
    dynamic = crs.datum.to_json_dict()["type"] == "DynamicGeodeticReferenceFrame"
    epoch = crs.datum.to_json_dict().get("frame_reference_epoch")
    return str(epoch) if dynamic and epoch else None


def add_epoch_option(s_crs: pp.CRS, t_crs: pp.CRS, warp_kwargs: dict):
    """
    Add epoch info to the GDAL warp options if either source or target CRSs are dynamic.

    Parameters
    ----------
    s_crs: pyproj.crs.CRS
        Source CRS object used in gdal Warp.
    t_crs: pyproj.crs.CRS
        Target CRS object used in gdal Warp.
    warp_kwargs: dict
        Optional GDAL warp options.

    Raises
    -------
    TypeError:
        If either one of `crs_from` or `crs_to` is not of type pyproj.CRS.

    Returns
    -------
    dict
        GDAL warp options.
    """
    if not isinstance(s_crs, pp.CRS) or not isinstance(t_crs, pp.CRS):
        raise TypeError("Both `s_crs` and `t_crs` must be of type pyproj.CRS.")
    s_epoch = crs_epoch(s_crs)
    t_epoch = crs_epoch(t_crs)
    options = {"options": []}
    if s_epoch:
        options["options"].append(f"s_coord_epoch={s_epoch}")
    if t_epoch:
        options["options"].append(f"t_coord_epoch={t_epoch}")
    if len(options["options"]) > 0:
        warp_kwargs |= {"options": options["options"]}
    return warp_kwargs


def validate_transform_steps(steps: Optional[list[str]]) -> bool:
    """
    Check if all transformation steps can be successfully instantiated by PROJ.

    Parameters
    ---------
    steps: Optional[list[str]]
        List of strings in form of `authority:code` representing the CRSs involved
        in a transformation pipeline.

    Raises
    -------
    NotImplementedError:
        When no transformer is identified.

    Returns
    --------
    bool:
        `False` if either one of the transformation steps fail, otherwise return `True`.
    """
    approve = True
    if not steps or len(steps) < 2:
        logger.error(f"{Fore.RED}Invalid transformation steps: {steps}")
        print(Style.RESET_ALL)
        return False
    for i in range(len(steps)-1):
        try:
            t1 = pp.Transformer.from_crs(crs_from=steps[i],
                                         crs_to=steps[i+1],
                                         allow_ballpark=False,
                                         only_best=True
                                         )
            tg = TransformerGroup(crs_from=steps[i],
                                  crs_to=steps[i+1],
                                  allow_ballpark=False,
                                  )
            # pyproj doesn't return proj string when there are more than 1 transformers
            if len(tg.transformers) < 2:
                if len(tg.transformers) == 0:
                    err_msg = (f"{Fore.RED}No transformers identified for the following "
                               f"transformation:\n\tcrs_from: {steps[i]}\n\tcrs_to: {steps[i+1]}")
                    logger.error(err_msg)
                    print(Style.RESET_ALL)
                    raise NotImplementedError(err_msg)
                ps = t1.to_proj4()
                error_hint = ""
                if not ps:
                    error_hint = "Null Proj string"
                elif "+proj=noop" in ps:
                    error_hint = "+proj=noop"
                elif "Error" in ps:
                    error_hint = "Error in Proj string"
                if error_hint:
                    logger.error(f"{Fore.RED}Invalid transformation step ({error_hint}): "
                                 f"{steps[i]} --> {steps[i+1]}")
                    print(Style.RESET_ALL)
                    approve = False
        except Exception as e:
            logger.error(f"{Fore.RED}Error in validation of transformation step: "
                         f"{steps[i]} --> {steps[i+1]}\n Error Msg: {e}",
                         stack_info=False, exc_info=False
                         )
            print(Style.RESET_ALL)
            approve = False
    return approve


def commandline(command: str,
                args: Optional[list[str]] = None) -> tuple[Optional[dict], Optional[str]]:
    """
    Spawn a new process to run a commandline utility and capture its output.

    Parameters
    -----------
    command: str
        The name of command (utility) to run. Example: `projinfo`
    args: Optional[list[str]]
        Optional arguments.

    Returns
    --------
    stdout: Optional[dict], std_err: Optional[str]
        standard output and error.
    """
    try:
        sout, serr = dict({}), None
        resp = subprocess.run([command, *args],
                              stderr=subprocess.PIPE,
                              stdout=subprocess.PIPE
                              )
        sout = resp.stdout.decode() if resp.stdout else None
        serr = resp.stderr.decode() if resp.stderr else None
    except Exception as e:
        logger.exception(str(e))
        sout, serr = dict({}), None
    return sout, serr


def validate_transform_steps_dict(steps: Optional[list[dict]]) -> bool:
    """
    Check if all transformation steps can be successfully instantiated by PROJ.

    Parameters
    ---------
    steps: Optional[list[dict]]
        List of dict objects containing crs_from/to in form of `authority:code`
        representing the CRSs involved in a transformation pipeline.

    Raises
    -------
    NotImplementedError:
        When no transformer is identified.

    Returns
    --------
    bool:
        `False` if either one of the transformation steps fail, otherwise return `True`.
    """
    approve = True
    if not steps or len(steps) < 1:
        logger.error(f"{Fore.RED}Invalid transformation steps: {steps}")
        print(Style.RESET_ALL)
        return False
    for i in range(len(steps)):
        try:
            t1 = pp.Transformer.from_crs(crs_from=steps[i]["crs_from"],
                                         crs_to=steps[i]["crs_to"],
                                         allow_ballpark=False,
                                         only_best=True
                                         )
            tg = TransformerGroup(crs_from=steps[i]["crs_from"],
                                  crs_to=steps[i]["crs_to"],
                                  allow_ballpark=False,
                                  )
            # pyproj doesn't return proj string when there are more than 1 transformers
            if len(tg.transformers) < 2:               
                if len(tg.transformers) == 0:
                    err_msg = (f"{Fore.RED}No transformers identified for the following "
                               f"transformation:\n\tcrs_from: {steps[i]['crs_from']}\n\tcrs_to: {steps[i]['crs_to']}")
                    logger.error(err_msg)
                    print(Style.RESET_ALL)
                    raise NotImplementedError(err_msg)
                ps = t1.to_proj4()
                error_hint = ""
                if not ps:
                    error_hint = "Null Proj string"
                elif "+proj=noop" in ps:
                    error_hint = "+proj=noop"
                elif "Error" in ps:
                    error_hint = "Error in Proj string"
                if error_hint:
                    logger.error(f"{Fore.RED}Invalid transformation step ({error_hint}): "
                                 f"{steps[i]['crs_from']} --> {steps[i]['crs_to']}")
                    print(Style.RESET_ALL)
                    approve = False
        except Exception as e:
            logger.error(f"{Fore.RED}Error in validation of transformation step: "
                         f"{steps[i]['crs_from']} --> {steps[i]['crs_to']}\n Error Msg: {e}",
                         stack_info=False, exc_info=False
                         )
            print(Style.RESET_ALL)
            approve = False
    return approve

def multiple_geodetic_crs(steps: Optional[list[dict]]) -> bool:
    """
    Check if there are more than one geodetic crs in the pipeline.

    Parameters
    ---------
    steps: Optional[list[dict]]
        List of dict objects containing crs_from/to in form of `authority:code`
        representing the CRSs involved in a transformation pipeline.


    Returns
    --------
    bool:
        `False` if all horizontal CRSs in the pipeline share the same geodetic CRS, otherwise return `True`.
    """
    geodetics = []
    for step in steps:
        h = step["crs_from"].split("+")[0]
        geodetics.append(":".join(pp.CRS(pp.CRS(h).geodetic_crs.to_2d()).to_authority()))
        h = step["crs_to"].split("+")[0]
        geodetics.append(":".join(pp.CRS(pp.CRS(h).geodetic_crs.to_2d()).to_authority()))
    return len(set(geodetics)) != 1

def multiple_projections(steps: Optional[list[dict]]) -> bool:
    """
    Check if there are more than one projection types in the pipeline.

    Parameters
    ---------
    steps: Optional[list[dict]]
        List of dict objects containing crs_from/to in form of `authority:code`
        representing the CRSs involved in a transformation pipeline.


    Returns
    --------
    bool:
        `False` if all projected CRSs use the same projection, otherwise return `True`.
    """
    def _projection_type(c: pp.CRS) -> str:
        lower_name = c.name.lower()
        spcs_keywords = ["state plane", "spcs", "fips", "stateplane"]
        if "utm" in lower_name:
            return "utm"
        elif any(s in lower_name for s in spcs_keywords):    
            return "spcs"
        else:
            return "others"
    projs = []
    for step in steps:
        c = pp.CRS(step["crs_from"].split("+")[0])
        if c.is_projected:
            projs.append(_projection_type(c))
        c = pp.CRS(step["crs_to"].split("+")[0])
        if c.is_projected:
            projs.append(_projection_type(c))
    return len(set(projs)) > 1

def pipeline_string(crs_from: str, crs_to, input_metadata=None) -> Optional[str]:
    """
    Extract PROJ pipeline string from the output of projinfo utility.

    Parameters
    -----------
    crs_from: str
        Source CRS in auth:code format.
    crs_to: str
        Target CRS in auth:code format.
    input_metadata: dict
        Input raster metadata object.

    Returns
    --------
    Optional[str]
    """
    args = ["-s", crs_from, "-t", crs_to,
            "--spatial-test", "intersects",
            "--hide-ballpark"]
    if input_metadata:
        bbox = [str(v) for v in input_metadata["geo_extent"]]
        args += ["--bbox", ",".join(bbox)]
    out, err = commandline(command="projinfo", args=args)
    if not out:
        raise ValueError(f"Potential error in getting projinfo output: {err}")

    start = out.find("+proj=pipeline")
    if start == -1:
        logger.error("`PROJ string:` not found in the projinfo output")
        return None

    splits = out[start:].splitlines(keepends=False)
    pipe = ""
    for split in splits:
        if len(split.strip()) == 0:
            break
        pipe += split
    return pipe



import os
import json
from vyperdatum.transformer import Transformer
from vyperdatum.utils.raster_utils import raster_metadata, update_raster_wkt
from vyperdatum.utils.vdatum_rest_utils import vdatum_cross_validate
import pyproj as pp
from pathlib import Path


def get_tiff_files(parent_dir: str, extention: str) -> list:
    tiff_files = []
    for (dirpath, dirnames, filenames) in os.walk(parent_dir):
        for filename in filenames:
            if filename.endswith(extention):
                tiff_files.append(os.sep.join([dirpath, filename]))
    return tiff_files



def transform_with_vyperdatum(input_file, output_file, crs_from, crs_to):
    steps = [
                {"crs_from": "EPSG:6347", "crs_to": "EPSG:6318", "v_shift": False},
                {"crs_from": "EPSG:6319", "crs_to": "EPSG:6318+NOAA:98", "v_shift": True},
                {"crs_from": "EPSG:6318+NOAA:98", "crs_to": "EPSG:6347+NOAA:98", "v_shift": False}
            ]

    tf = Transformer(crs_from=crs_from,
                     crs_to=crs_to,
                    #  steps=steps
                     )
    
    tf.transform_raster(input_file=input_file,
                        output_file=output_file,
                        overview=False,
                        pre_post_checks=True,
                        vdatum_check=False
                        )
    return


if __name__ == "__main__":
    parent_dir = r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\HRD\Original"
    files = get_tiff_files(parent_dir, extention=".tif")


    crs_from = "EPSG:32618+NOAA:86"
    crs_to = "EPSG:32618+EPSG:5703"

    # files = [r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\NC\Original\NC1903-TB-C_BLK-07\NC1903-TB-C_BLK-07_US4NC1FG_ellipsoidal_dem.tif"]

    for i, input_file in enumerate(files[:1]):
        try:
            print(f"{i+1}/{len(files)}: {input_file}")
            output_file = input_file.replace("Original", "Manual")
            transform_with_vyperdatum(input_file, output_file, crs_from, crs_to)
            print(f'\n{"*"*50} {i+1}/{len(files)} Completed {"*"*50}\n')

            meta = raster_metadata(output_file)
            print(json.dumps(meta["Vyperdatum_Metadata"], sort_keys=True, indent=4))
        except Exception as e:
                efile = open(Path(output_file).parent.absolute()/Path(f"{os.path.split(input_file)[1]}_error.txt"), "w")
                efile.write(str(e))
                efile.close()




import os
from pathlib import Path
from typing import Optional, List
import numpy as np
import pandas as pd
import geopandas as gpd
from shapely.geometry import Point
from vyperdatum.drivers.base import Driver
import pyproj as pp
import logging
from osgeo import gdal
from vyperdatum.utils.crs_utils import auth_code
from vyperdatum.drivers.base import Driver


logger = logging.getLogger("root_logger")
gdal.UseExceptions()

class XYZ(Driver):
    """
    Handle loading and parsing of a generic .xyz.

    Attributes
    ----------
    input_file: str
        Full path to the xyz file.
    invalid_error: bool = True
        If True, throws an error when the input file has an unexpected format.
    skiprows: Optional[int] = None
        The number of rows at the beginning of the file to ignore.
    col_names: Optional[List[str]] = None
        The names of the columns in the file. If None, defaults to ['x', 'y', 'z'].
    negate_z: bool = False
        If True, negates the z values in the file.
    unit_conversion: float = 1.0
        A factor to convert the z values to the desired unit.

    Pseudo Example
    --------
    >>> xyz = XYZZ('PATH_TO_NPZ_FILE')
    >>> df = xyz.transform(crs_from=crs_from, crs_to=crs_to, steps=steps)
    >>> print(df.head())
    >>> xyz.to_gpkg(crs=crs_to, output_file=output_file)
    """

    def __init__(self, input_file: str,
                 invalid_error: bool = True,
                 skiprows: Optional[int] = None,
                 col_names: Optional[List[str]] = None,
                 negate_z: bool = False,
                 unit_conversion: float = 1.0
                 ) -> None:
        """
        Load a numpy .xyz file (a csv file with generic comments at the top).

        Parameters
        ----------
        input_file: str
            Full file path.
        invalid_error: bool, default True
            If True, throws an error when the input file has an unexpected format.

        Raises
        --------
        ValueError:
            If the input file is not recognized as xyz file.

        Returns
        -----------
        None
        """
        super().__init__()
        self.input_file = input_file
        try:
            self.skiprows = self._detect_data_start() if skiprows is None else skiprows
            self.col_names = col_names
            self.negate_z = negate_z
            self.unit_conversion = unit_conversion
            self.df = self._parse()
            self.is_xyz = True
        except:
            self.is_xyz = False
        if invalid_error and not self.is_xyz:
            raise ValueError(f"Not a valid .xyz file: {self.input_file}.")

    def _detect_data_start(self) -> Optional[int]:
        with open(self.input_file, "r") as f:
            for i, line in enumerate(f):
                parts = line.strip().split(",")
                if len(parts) >= 3:
                    try:
                        float(parts[0])
                        float(parts[1])
                        float(parts[2])
                        return i
                    except ValueError:
                        continue
        raise ValueError("No valid data lines found in file.")

    def _parse(self) -> pd.DataFrame:
        """
        Reads the .XYZ file into a pandas DataFrame.
        If self.col_names is not set, assumes the first three columns are x, y, z.

        Returns:
            pd.DataFrame: the parsed data
        """
        df = pd.read_csv(self.input_file, sep=",", skiprows=self.skiprows)
        num_cols = df.shape[1]
        if self.col_names:
            base_names = self.col_names
        else:
            base_names = ["x", "y", "z"]
        if num_cols > len(base_names):
            column_names = base_names + [f"col{i}" for i in range(len(base_names)+1, num_cols + 1)]
        else:
            column_names = base_names[:num_cols]
        df.columns = column_names
        if self.negate_z:
            df["z"] = -df["z"]
        df["z"] *= self.unit_conversion
        return df

    def wkt(self) -> str:
        """
        Return wkt stored in the xyz file.
        Will raise NotImplementedError as xyz files do not have a dedicated attribute for WKT.
        """
        raise NotImplementedError("xyz files have no dedicated attribute for WKT.")

    def transform(self, transformer_instance, output_file: str, pre_post_checks: bool, vdatum_check: bool) -> bool:
        """
        Apply point transformation on npz data according to the `transformer_instance`.

        Parameters
        -----------
        transformer_instance: vyperdatum.transformer.Transform
            Instance of the transformer class.
        output_file: str
            The output file path where the transformed data will be saved.
        vdatum_check: bool
            If True, performs vertical datum checks after transformation.

        Returns
        -----------
        bool:
            True if successful, otherwise False.
        """
        x, y, z = self.df["x"].values, self.df["y"].values, self.df["z"].values

        if pre_post_checks:
            # Cannot check the CRS of xyz file as it has no wkt.
            pass
            # source_crs = transformer_instance.crs_from
            # if not isinstance(transformer_instance.crs_from, pp.CRS):
            #     source_crs = pp.CRS(transformer_instance.crs_from)
            # source_auth = auth_code(source_crs)
            # file_auth = auth_code(pp.CRS(self.wkt()))
            # if source_auth != file_auth:
            #     logger.warning("The authority name/code registered in the "
            #                    f"input file is {file_auth}, but expected {source_auth}"
            #                    )

        success, xt, yt, zt = transformer_instance.transform_points(x, y, z,
                                                                    vdatum_check=vdatum_check)
        if not success:
            raise ValueError(f"XYZ transformation failed for {self.input_file}.")
        self.df["x_t"], self.df["y_t"], self.df["z_t"] = xt, yt, zt
        if self.negate_z:
            self.df["z"] = -self.df["z"]
            self.df["z_t"] = -self.df["z_t"]
        self.df["Uncertainty"] = 1 + 0.02 * self.df["z_t"].abs()

        # output_file = self.to_gpkg(crs=transformer_instance.crs_to, output_file=output_file)
        _ = self.to_npy(output_file=output_file)

        # if pre_post_checks:
        #     new_gdf = gpd.read_file(output_file)
        #     target_crs = transformer_instance.crs_to
        #     if not isinstance(transformer_instance.crs_to, pp.CRS):
        #         target_crs = pp.CRS(transformer_instance.crs_to)
        #     target_auth = auth_code(target_crs)
        #     transformed_file_auth = auth_code(pp.CRS(new_gdf.crs.to_wkt()))
        #     if target_auth != transformed_file_auth:
        #         logger.warning("The expected authority name/code of the "
        #                        f"transformed geoparquet is {target_auth}, but received {transformed_file_auth}"
        #                        )
        return success

    def to_npy(self,
               output_file: str) -> None:
        try:
            output_file = str(Path(output_file).with_suffix(".npy"))
            out_npy = np.stack((self.df["x_t"].values,
                                self.df["y_t"].values,
                                self.df["z_t"].values,
                                self.df["Uncertainty"].values
                                ), axis=1)
            np.save(output_file, out_npy)
        except Exception as e:
            if os.path.isfile(output_file):
                os.remove(output_file)
            raise RuntimeError(f"Failed to write output to npy: {output_file}. Error: {e}")
        return output_file


    def to_gpkg(self,
                crs: str,
                output_file: str) -> None:
        try:
            output_file = str(Path(output_file).with_suffix(".gpkg"))
            tdf = pd.DataFrame({"x": self.df["x_t"].values,
                                "y": self.df["y_t"].values,
                                "Elevation": self.df["z_t"].values,
                                })
            tdf["geometry"] = tdf.apply(lambda row: Point(row["x"], row["y"], row["Elevation"]), axis=1)
            tdf["Uncertainty"] = self.df["Uncertainty"].values
            gdf = gpd.GeoDataFrame(tdf, geometry="geometry", crs=crs)
            gdf.to_file(output_file, driver="GPKG")
        except Exception as e:
            if os.path.isfile(output_file):
                os.remove(output_file)
            raise RuntimeError(f"Failed to write output to GPKG: {output_file}. Error: {e}")
        return output_file

    @property
    def is_valid(self):
        return self.is_xyz



import os
import platform
import sqlite3
import logging
from typing import Optional, Tuple, Union
import pyproj as pp
from pathlib import Path
import pandas as pd
from vyperdatum.enums import PROJDB as enuPDB


logger = logging.getLogger("root_logger")


class DB:

    def __init__(self,
                 db_dir: Optional[str] = enuPDB.DIR.value
                 ) -> None:
        """

        Parameters
        ----------
        db_dir: str
            Path to the directory where the proj database file is located. The
            default values is pyproj data directory (`pyproj.datadir`).

        Returns
        -------
        NoneType
        """
        self.db_dir = db_dir
        self.db_name = enuPDB.FILE_NAME.value
        self.db_file_path = Path(self.db_dir).joinpath(self.db_name)

        if db_dir != pp.datadir.get_data_dir():
            self.update_db_path()
        return

    def __str__(self):
        return (f"db_dir: {self.db_dir}\ndb_file_path: {self.db_file_path}"
                f"\nproj.datadir: {pp.datadir.get_data_dir()}"
                )

    def __repr__(self):
        return f"DB(data_dir = r'{self.db_dir}')"

    @property
    def db_file_path(self) -> str:
        return self._db_file_path

    @db_file_path.setter
    def db_file_path(self, db_path: str):
        if not db_path.is_file():
            raise FileNotFoundError(f"Proj Database file not found at: {db_path}")
        self.db_dir = os.path.dirname(db_path)
        self.db_name = os.path.basename(db_path)
        self._db_file_path = db_path

    def update_db_path(self) -> bool:
        """
        Prepend `self.db_dir` to `pyproj.datadir` which guides the pyproj
        to first look for the database at `self.db_dir` address.

        Raises
        -------
        ValueError:
            If `.db_dir` is not set.
        FileNotFoundError:
            If the database file is not found.


        Returns
        -------
        bool:
            `True` if the `data_dir` is set successfully, otherwise `False`.
        """
        try:
            success = False
            if not self.db_dir:
                raise ValueError("Attribute `.db_dir` not specified.")
            sep = ";" if platform.system() == "Windows" else ":"
            pp.datadir.set_data_dir(self.db_dir + sep + pp.datadir.get_data_dir())
            if pp.datadir.get_data_dir().split(sep)[0] != self.db_dir:
                raise SystemExit("Unable to set the path to the custom PROJ database.")
            else:
                success = True
        finally:
            return success

    def _connect(self) -> Tuple[Optional[object], Optional[object]]:
        """
        Connect to the proj database.

        Returns
        -----------
        obj:
            database connection object
        obj:
            database cursor object
        """
        try:
            con, cur = None, None
            con = sqlite3.connect(self.db_file_path)
            cur = con.cursor()
        except Exception:
            logger.exception(f"Error while connecting to database at {self.db_file_path}.")
        return con, cur

    def query(self,
              sql: str,
              dataframe: bool = False
              ) -> Union[Optional[list], Optional[pd.DataFrame]]:
        """
        Execute a sql query and return the response. This method is intended
        to run a read (scan) query.
        Avoid using this method for DML/DDL type operations.

        Parameters
        ----------
        sql: str
            SQL query (intended to be a scan query) to be executed.
        dataframe: bool, default=False
            If True, converts the result into a pandas dataframe.

        Returns
        --------
        list or pd.DataFrame
        """
        try:
            con, cur, res = None, None, None
            con, cur = self._connect()
            cur.execute(sql)
            res = cur.fetchall()
            # logger.info(res)
            if dataframe:
                res = pd.DataFrame.from_records(res,
                                                columns=[column[0] for column in cur.description]
                                                )
        except Exception:
            logger.exception("Error in db.query.")
        finally:
            if cur:
                cur.close()
            if con:
                con.close()
        return res

    def crs_by_keyword(self,
                       keywords: list[str],
                       dataframe: bool = False
                       ) -> Union[Optional[list], Optional[pd.DataFrame]]:
        """
        Return a list (or dataframe) of CRS that their name or description
        contain the passed keywords. The search is not case-sensitive.

        Parameters
        ----------
        sql: str
            SQL query (intended to be a scan query) to be executed.
        keywords: list[str]
            A list of string keywords used to query the database.
        dataframe: bool, default=False
            If True, converts the result into a pandas dataframe.

        Raises
        -------
        TypeError:
            If `keywords` is not a list of strings.
        ValueError:
            If no keywords is passed.

        Returns
        --------
        list or pd.DataFrame
        """
        if not isinstance(keywords, list):
            raise TypeError("keywords must be a list")
        if not all(map(lambda k: isinstance(k, str), keywords)):
            raise TypeError("keywords must be a list of strings")
        if len(keywords) < 1:
            raise ValueError("at least one keyword most be passed")
        where = ""
        filters = [f"(name like '%{k}%' OR description like '%{k}%')" for k in keywords]
        if len(filters) > 0:
            where = "WHERE " + " AND ".join(filters)
        return self.query(sql=f"SELECT * FROM {enuPDB.VIEW_CRS.value} {where}",
                          dataframe=dataframe)



import os
import glob
import pathlib
import pandas as pd
from vyperdatum.transformer import Transformer
import geopandas as gpd
from shapely.geometry import Point


if __name__ == "__main__":
    steps = [
            {"crs_from": "ESRI:102684", "crs_to": "EPSG:4267"},
            {"crs_from": "EPSG:4267+EPSG:5702", "crs_to": "EPSG:6318+NOAA:98"},
            {"crs_from": "EPSG:6318", "crs_to": "EPSG:6348"}
            ]
    input_file = r"C:\Users\mohammad.ashkezari\Desktop\BASIN 3-27-24.csv"
    temp_file = r"C:\Users\mohammad.ashkezari\Desktop\temp.csv"
    output_file = r"C:\Users\mohammad.ashkezari\Desktop\BASIN.csv"
    df = pd.read_csv(input_file)
    df["Raw Elevation Reading"] = -1 * df["Raw Elevation Reading"]
    x, y, z = df["Easting"].values, df["Northing"].values, df["Raw Elevation Reading"].values
    tf = Transformer(crs_from=steps[0]["crs_from"],
                     crs_to=steps[-1]["crs_to"],
                     steps=steps)
    success, xt, yt, zt = tf.transform_points(x, y, z,
                                        always_xy=True,
                                        allow_ballpark=False,
                                        only_best=True,
                                        vdatum_check=False)
    df["x_t"], df["y_t"] = xt, yt
    df["z_t"] = zt
    df.to_csv(output_file, index=False)

    tdf = pd.DataFrame({"x": xt, "y": yt, "z": zt})
    tdf.to_csv(temp_file, index=False)
    tdf["geometry"] = tdf.apply(lambda row: Point(row["x"], row["y"], row["z"]), axis=1)
    gdf = gpd.GeoDataFrame(tdf, geometry="geometry", crs="EPSG:6348+NOAA:98")
    gdf.to_file(output_file.replace("csv", "gpkg"), driver="GPKG")



import glob
from vyperdatum.transformer import Transformer
from vyperdatum.utils.raster_utils import raster_metadata, update_raster_wkt
from vyperdatum.utils.vdatum_rest_utils import vdatum_cross_validate
import pyproj as pp




if __name__ == "__main__":

    parent_dir = r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\PBA\PBA_WestCoast_UTM10N_NCD\Original\**\*.bag"
    files = glob.glob(parent_dir, recursive=True)[:1]
    crs_from = "EPSG:26910+NOAA:98"
    crs_to = "EPSG:26910+NOAA:88"

    # crs_from = "EPSG:26910+EPSG:5866"
    # crs_to = "EPSG:6339+EPSG:5866"
    # steps = [{'crs_from': 'EPSG:26910', 'crs_to': 'EPSG:6318', 'v_shift': False},
    #          {'crs_from': 'EPSG:6318', 'crs_to': 'EPSG:6339', 'v_shift': False}]

    steps = [
        {'crs_from': 'EPSG:26910', 'crs_to': 'EPSG:6318', 'v_shift': False},
        {'crs_from': 'EPSG:6318+NOAA:98', 'crs_to': 'EPSG:6318+NOAA:88', 'v_shift': True},
        # {'crs_from': 'EPSG:6318', 'crs_to': 'EPSG:26910', 'v_shift': False},
             ]


    # ###################################################
    # import pytest
    # import sys
    # # coords_lat_lon = (39, -76.5, 0)
    
    # # steps = [
    # # {'crs_from': 'EPSG:6349', 'crs_to': 'EPSG:6319', 'v_shift': True}
    # #         ] 

    # # success, x, y, z = Transformer(crs_from="EPSG:6349", crs_to="EPSG:6319", steps=steps).transform_points([coords_lat_lon[0]],[coords_lat_lon[1]],[coords_lat_lon[2]])
    # # print(x, y, z)
    # # # assert pytest.approx(x, abs=.01) == coords_lat_lon[0], "x coordinate should remain unchanged."
    # # # assert pytest.approx(y, abs=.01) == coords_lat_lon[1], "y coordinate should remain unchanged."
    # # # assert pytest.approx(z, abs=.01) == -33.29, "incorrect z coordinate transformation."    
    # # sys.exit



    # x1 = 613922.931016
    # y1 = 5061381.271016

    # ## fake inside
    # x1 = 614956
    # y1 = 5064834

    # x2 = 618970.618984
    # y2 = 5066428.958984

    # tf = Transformer(crs_from=crs_from,
    #                     crs_to=crs_to,
    #                     steps=steps
    #                     )
    # # _, y1, x1, _ = tf.transform_points([y1], [x1], [0], always_xy=False, allow_ballpark=False)
    # # _, y2, x2, _ = tf.transform_points([y2], [x2], [0], always_xy=False, allow_ballpark=False)

    # _, xx, yy, zz = tf.transform_points([y1], [x1], [0])
    # print(xx, yy, zz)

    
    # sys.exit()
    # ###################################################


    for i, input_file in enumerate(files):
        print(f"{i+1}/{len(files)}: {input_file}")
        # meta = raster_metadata(input_file)
        # print(meta['h_authcode'], meta['v_authcode'])
        tf = Transformer(crs_from=crs_from,
                         crs_to=crs_to,
                         steps=steps
                         )
        output_file = input_file.replace("Original", "Manual")
        tf.transform(input_file=input_file,
                     output_file=output_file,
                     pre_post_checks=True,
                     vdatum_check=False
                     )
        print(f'\n{"*"*50} {i+1}/{len(files)} Completed {"*"*50}\n')



import logging
from osgeo import gdal
from typing import Union
import pyproj as pp
from vyperdatum.drivers import vrbag
from vyperdatum.drivers.laz import LAZ
from vyperdatum.drivers.npz import NPZ
from vyperdatum.drivers.pdal_based import PDAL
from vyperdatum.utils.crs_utils import auth_code


logger = logging.getLogger("root_logger")
gdal.UseExceptions()


def vrbag_pre_transformation_checks(file_path: str,
                                    source_crs: Union[pp.CRS, str]
                                    ) -> bool:
    """
    Run a number of sanity checks on the source file before transformation.
    Warns if a check fails.

    Parameters
    ----------
    file_path: str
        Path to the input file.
    source_crs: pyproj.crs.CRS or input used to create one
        The expected CRS object for the vrbag file.

    Returns
    ----------
    bool
        Returns True if all checks pass, otherwise False.
    """
    passed = True
    if not isinstance(source_crs, pp.CRS):
        source_crs = pp.CRS(source_crs)
    source_auth = auth_code(source_crs)
    file_auth = auth_code(pp.CRS(vrbag.wkt(fname=file_path)))
    if source_auth != file_auth:
        passed = False
        logger.warning("The authority name/code registered in the "
                       f"input file is {file_auth}, but received {source_auth}"
                       )
    if not vrbag.is_vr(file_path):
        passed = False
        logger.warning(f"The input file ({file_path}) is not a valid Vrbag file.")
    return passed


def vrbag_post_transformation_checks(file_path: str,
                                     target_crs: Union[pp.CRS, str]
                                     ) -> bool:
    """
    Run a number of sanity checks on the transformed vrbag file.
    Warns if a check fails.

    Parameters
    ----------
    file_path: str
        Path to the input file.
    target_crs: pyproj.crs.CRS or input used to create one
        The expected CRS object for the transformed file.

    Returns
    ----------
    bool
        Returns True if all checks pass, otherwise False.
    """
    if not isinstance(target_crs, pp.CRS):
        target_crs = pp.CRS(target_crs)
    passed = True
    target_auth = auth_code(target_crs)
    transformed_file_auth = auth_code(pp.CRS(vrbag.wkt(fname=file_path)))
    if target_auth != transformed_file_auth:
        passed = False
        logger.warning("The expected authority name/code of the "
                       f"transformed vrbag is {target_auth}, but received {transformed_file_auth}"
                       )
    if not vrbag.is_vr(file_path):
        passed = False
        logger.warning(f"The transformed file ({file_path}) is not a valid Vrbag file.")
    return passed


def laz_pre_transformation_checks(file_path: str,
                                  source_crs: Union[pp.CRS, str]
                                  ) -> bool:
    """
    Run a number of sanity checks on the source file before transformation.
    Warns if a check fails.

    Parameters
    ----------
    file_path: str
        Path to the input file.
    source_crs: pyproj.crs.CRS or input used to create one
        The expected CRS object for the laz file.

    Returns
    ----------
    bool
        Returns True if all checks pass, otherwise False.
    """
    passed = True
    if not isinstance(source_crs, pp.CRS):
        source_crs = pp.CRS(source_crs)
    source_auth = auth_code(source_crs)
    laz = LAZ(input_file=file_path)
    file_auth = auth_code(pp.CRS(laz.wkt()))
    if source_auth != file_auth:
        passed = False
        logger.warning("The authority name/code registered in the "
                       f"input file is {file_auth}, but received {source_auth}"
                       )
    return passed


def laz_post_transformation_checks(file_path: str,
                                   target_crs: Union[pp.CRS, str]
                                   ) -> bool:
    """
    Run a number of sanity checks on the transformed laz file.
    Warns if a check fails.

    Parameters
    ----------
    file_path: str
        Path to the input file.
    target_crs: pyproj.crs.CRS or input used to create one
        The expected CRS object for the transformed file.

    Returns
    ----------
    bool
        Returns True if all checks pass, otherwise False.
    """
    if not isinstance(target_crs, pp.CRS):
        target_crs = pp.CRS(target_crs)
    passed = True
    target_auth = auth_code(target_crs)
    laz = LAZ(input_file=file_path)
    transformed_file_auth = auth_code(pp.CRS(laz.wkt()))
    if target_auth != transformed_file_auth:
        passed = False
        logger.warning("The expected authority name/code of the "
                       f"transformed LAZ is {target_auth}, but received {transformed_file_auth}"
                       )
    return passed


def npz_pre_transformation_checks(file_path: str,
                                  source_crs: Union[pp.CRS, str]
                                  ) -> bool:
    """
    Run a number of sanity checks on the source file before transformation.
    Warns if a check fails.

    Parameters
    ----------
    file_path: str
        Path to the input file.
    source_crs: pyproj.crs.CRS or input used to create one
        The expected CRS object for the npz file.

    Returns
    ----------
    bool
        Returns True if all checks pass, otherwise False.
    """
    passed = True
    if not isinstance(source_crs, pp.CRS):
        source_crs = pp.CRS(source_crs)
    source_auth = auth_code(source_crs)
    nz = NPZ(input_file=file_path)
    file_auth = auth_code(pp.CRS(nz.wkt()))
    if source_auth != file_auth:
        passed = False
        logger.warning("The authority name/code registered in the "
                       f"input file is {file_auth}, but received {source_auth}"
                       )
    return passed


def npz_post_transformation_checks(file_path: str,
                                   target_crs: Union[pp.CRS, str]
                                   ) -> bool:
    """
    Run a number of sanity checks on the transformed npz file.
    Warns if a check fails.

    Parameters
    ----------
    file_path: str
        Path to the input file.
    target_crs: pyproj.crs.CRS or input used to create one
        The expected CRS object for the transformed file.

    Returns
    ----------
    bool
        Returns True if all checks pass, otherwise False.
    """
    if not isinstance(target_crs, pp.CRS):
        target_crs = pp.CRS(target_crs)
    passed = True
    target_auth = auth_code(target_crs)
    nz = NPZ(input_file=file_path)
    transformed_file_auth = auth_code(pp.CRS(nz.wkt()))
    if target_auth != transformed_file_auth:
        passed = False
        logger.warning("The expected authority name/code of the "
                       f"transformed NPZ is {target_auth}, but received {transformed_file_auth}"
                       )
    return passed


def pdal_pre_transformation_checks(file_path: str,
                                   source_crs: Union[pp.CRS, str]
                                   ) -> bool:
    """
    Run a number of sanity checks on the source file before transformation.
    Warns if a check fails.

    Parameters
    ----------
    file_path: str
        Path to the input file.
    source_crs: pyproj.crs.CRS or input used to create one
        The expected CRS object for the PDAL-supported file.

    Returns
    ----------
    bool
        Returns True if all checks pass, otherwise False.
    """
    passed = True
    if not isinstance(source_crs, pp.CRS):
        source_crs = pp.CRS(source_crs)
    source_auth = auth_code(source_crs)
    pdl = PDAL(input_file=file_path, output_file="")
    file_auth = auth_code(pp.CRS(pdl.wkt(fname=file_path)))
    if source_auth != file_auth:
        passed = False
        logger.warning("The authority name/code registered in the "
                       f"input file is {file_auth}, but received {source_auth}"
                       )
    return passed


def pdal_post_transformation_checks(file_path: str,
                                    target_crs: Union[pp.CRS, str]
                                    ) -> bool:
    """
    Run a number of sanity checks on the transformed PDAL-supported file.
    Warns if a check fails.

    Parameters
    ----------
    file_path: str
        Path to the input file.
    target_crs: pyproj.crs.CRS or input used to create one
        The expected CRS object for the transformed file.

    Returns
    ----------
    bool
        Returns True if all checks pass, otherwise False.
    """
    if not isinstance(target_crs, pp.CRS):
        target_crs = pp.CRS(target_crs)
    passed = True
    target_auth = auth_code(target_crs)
    pdl = PDAL(input_file=file_path, output_file="")
    transformed_file_auth = auth_code(pp.CRS(pdl.wkt(fname=file_path)))
    if target_auth != transformed_file_auth:
        passed = False
        logger.warning("The expected authority name/code of the "
                       f"transformed PDAL-supported file is {target_auth},"
                       f" but received {transformed_file_auth}"
                       )
    return passed



import os
import pathlib
import glob
from osgeo import gdal
from vyperdatum.transformer import Transformer
from vyperdatum.utils.raster_utils import raster_metadata


def transform_19n(input_file):
    """
    Transform from NAD83(2011) / UTM zone 19N + MLLW to NAD83(2011) / UTM zone 19N + PRVD
    """
    # Currently for Puerto Rico there is no "NAD83(2011) + MLLW" CRS is defined in the database. 
    # The closest CRS I can find is "NAD83(2011) + MSL (GEOID12B_PRVI) height": NOAA:8283 = EPSG:6318+NOAA:5535


    warp_kwargs_vertical = {
                            "outputType": gdal.gdalconst.GDT_Float32,
                            "srcBands": [1],
                            "dstBands": [1],
                            "warpOptions": ["APPLY_VERTICAL_SHIFT=YES", "SAMPLE_GRID=YES", "SAMPLE_STEPS=ALL"],
                            "errorThreshold": 0,
                            }
    
    # Horizontal: NAD83(2011) / UTM zone 19N + MLLW  >>>>  NAD83(2011) + MLLW
    t1 = Transformer(crs_from="EPSG:26919+NOAA:5535",
                     crs_to="NOAA:8283", # NOAA:8283 = EPSG:6318+NOAA:5535
                     allow_ballpark=False
                     )
    out_file1 = pathlib.Path(input_file).with_stem("_01_" + pathlib.Path(input_file).stem)
    t1.transform_raster(input_file=input_file,
                        output_file=out_file1,
                        apply_vertical=False
                        )

    # Vertical: NAD83(2011) + MSL (GEOID12B_PRVI) height >>>>  NAD83(2011) + PRVD02
    t2 = Transformer(crs_from="NOAA:8283", # NOAA:8283 = EPSG:6318+NOAA:5535
                     crs_to="EPSG:9522", #"EPSG:9522 = EPSG:6318+EPSG:6641
                     allow_ballpark=False  # have to set it to True (results in noop). Setting crs_to=NOAA:8552 also results in noop
                     )
    out_file2 = pathlib.Path(input_file).with_stem("_02_" + pathlib.Path(input_file).stem)
    t2.transform_raster(input_file=out_file1,
                        output_file=out_file2,
                        apply_vertical=True,
                        warp_kwargs=warp_kwargs_vertical
                        )

    # Project: NAD83(2011)  >>>>  NAD83(2011) / UTM zone 19N
    t3 = Transformer(crs_from="EPSG:9522", #"EPSG:9522 = EPSG:6318+EPSG:6641
                     crs_to="EPSG:26919+EPSG:6641",
                     allow_ballpark=False
                     )
    # out_file3 = pathlib.Path(input_file).with_stem("_03_" + pathlib.Path(input_file).stem)
    p = pathlib.Path(input_file)
    xform_dir = os.path.join(r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\PRVD\Manual", p.parent.parent.name, p.parent.name)
    os.makedirs(xform_dir, exist_ok=True)
    out_file3 = os.path.join(xform_dir, p.name)

    t3.transform_raster(input_file=out_file2,
                        output_file=out_file3,
                        apply_vertical=False
                        )

    os.remove(out_file1)
    os.remove(out_file2)
    return


if __name__ == "__main__":
    # files = get_tiff_files(r"W:\working_space\test_environments\sandbox\PBG19") + get_tiff_files(r"W:\working_space\test_environments\sandbox\PBG20")
    files = glob.glob(r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\PRVD\Original\**\**\*.tiff", recursive=True)
    for i, input_file in enumerate(files):
        print(f"{i+1}/{len(files)}: {input_file}")
        transform_19n(input_file)



from vyperdatum.transformer import Transformer
from vyperdatum.utils.raster_utils import raster_metadata

# input_file = r"C:\Users\mohammad.ashkezari\Desktop\E01001_MB_VR_MLLW_2of5.bag"
# output_file = r"C:\Users\mohammad.ashkezari\Desktop\E01001_MB_VR_NAVD88_2of5.bag"

# crs_from = "EPSG:6335+NOAA:98"
# crs_to = "EPSG:6335+EPSG:5703"
# tf = Transformer(crs_from=crs_from, crs_to=crs_to )                
# tf.transform_raster(input_file=input_file,
#                     output_file=output_file,
#                     overview=False,
#                     pre_post_checks=True,
#                     vdatum_check=False
#                     )




# https://vlab.noaa.gov/redmine/issues/148720
# ------------------------------------------------------------------------
# input_file = r"C:\Users\mohammad.ashkezari\Desktop\wgs\E01096_MB_2m_MLLW_1of7.bag"
# output_file = r"C:\Users\mohammad.ashkezari\Desktop\wgs\E01096_MB_2m_MLLW_1of7.bag"

# meta = raster_metadata(input_file)
# print(meta["h_authcode"])
# print(meta["v_authcode"])
# crs_from = "EPSG:6335+NOAA:98"
# crs_to = "EPSG:6335+EPSG:5703"
# tf = Transformer(crs_from=crs_from, crs_to=crs_to )                
# tf.transform_raster(input_file=input_file,
#                     output_file=output_file,
#                     overview=False,
#                     pre_post_checks=True,
#                     vdatum_check=False
#                     )





from vyperdatum.transformer import Transformer
from io import BytesIO
import h5py
import numpy as np
from lxml import etree
import pyproj as pp


def transform_w(p):
    crs_from = "EPSG:32617+EPSG:5866"
    crs_to = "EPSG:26917+EPSG:5866"
    tf = Transformer(crs_from=crs_from,
                     crs_to=crs_to,
                     steps=["EPSG:32617", "EPSG:9755", "EPSG:6318", "EPSG:26917"]
                     )
    success, y, x, z = tf.transform_points(p[1], p[0], 0, always_xy=False, allow_ballpark=False)
    return f"{x},{y}"


def transform_h(p):
    crs_from = "EPSG:32619+EPSG:5866"
    crs_to = "EPSG:6348+EPSG:5866"
    tf = Transformer(crs_from=crs_from,
                     crs_to=crs_to,
                     steps=["EPSG:32619", "EPSG:9755", "EPSG:6318", "EPSG:6348"]
                     )
    success, y, x, z = tf.transform_points(p[1], p[0], 0, always_xy=False, allow_ballpark=False)
    return f"{x},{y}"


def corner_points(bag_fname: str):
    """
    Return the corner points.
    """
    bag = h5py.File(bag_fname)
    meta = bag["BAG_root/metadata"]
    buffer = BytesIO(meta[()])
    tree = etree.parse(buffer)
    root = tree.getroot()
    gml = ".//{" + root.nsmap['gml'] + "}"
    return root.find(f"{gml}coordinates").text


def change_corner_points(bag_fname: str, new_points: str, wkt_h: str, wkt_v: str):
    """
    Update the metadata's corner points with new points.
    """
    bag = h5py.File(bag_fname)
    meta = bag["BAG_root/metadata"]
    buffer = BytesIO(meta[()])
    tree = etree.parse(buffer)
    root = tree.getroot()
    gml = ".//{" + root.nsmap['gml'] + "}"
    gco = ".//{" + root.nsmap['gco'] + "}"
    root.find(f"{gml}coordinates").text = new_points
    root.findall(f"{gco}CharacterString")[6].text = wkt_h
    root.findall(f"{gco}CharacterString")[8].text = wkt_v
    # tree.write(xml_fname)
    # xml = etree.tostring(root, pretty_print=True).decode("ascii")
    xmet = etree.tostring(root).decode()
    bag.close()
    bag = h5py.File(bag_fname, mode="r+")
    root = bag.require_group("/BAG_root")
    del bag["/BAG_root/metadata"]
    metadata = np.array(list(xmet), dtype="S1")
    root.create_dataset("metadata", maxshape=(None,), data=metadata, compression="gzip", compression_opts=9)
    bag.close()
    return


def update_W00656():
    w_p1 = (278881.198477, 2719890.433477)
    w_p2 = (293586.271523, 2734595.506523)
    new_w_corners = transform_w(w_p1) + " " + transform_w(w_p2)
    bag_fname = "W00656_MB_VR_MLLW_5of5.bag"
    print(corner_points(bag_fname=bag_fname))
    change_corner_points(bag_fname=bag_fname, new_points=new_w_corners, wkt_h=pp.CRS("EPSG:26917").to_wkt(), wkt_v=pp.CRS("EPSG:5866").to_wkt())
    print(corner_points(bag_fname=bag_fname))
    return


def update_H12137():
    h_p1 = (259274.885634, 4549506.909634)
    h_p2 = (280112.069366, 4570344.093366)
    new_h_corners = transform_h(h_p1) + " " + transform_h(h_p2)
    bag_fname = "H12137_MB_VR_MLLW_1of1.bag"
    print(corner_points(bag_fname=bag_fname))
    change_corner_points(bag_fname=bag_fname, new_points=new_h_corners, wkt_h=pp.CRS("EPSG:6348").to_wkt(), wkt_v=pp.CRS("EPSG:5866").to_wkt())
    print(corner_points(bag_fname=bag_fname))
    return


update_W00656()
update_H12137()



# bag_fname = "W00656_MB_VR_MLLW_5of5.bag"
# bag = h5py.File(bag_fname)
# meta = bag["BAG_root/metadata"]
# buffer = BytesIO(meta[()])
# tree = etree.parse(buffer)
# root = tree.getroot()
# gco = ".//{" + root.nsmap['gco'] + "}"
# css = root.findall(f"{gco}CharacterString")
# print(len(css))
# for i, cs in enumerate(css):
#     print(i)
#     print(cs.text[:5])
#     print("///////")
# bag.close()




import pandas as pd
from vyperdatum.db import DB
import pyproj as pp
from pyproj.transformer import TransformerGroup


def test_noaa_transforms():
    """
    Check if PROJ can instantiate a transformer using the NOAA-defined transformations.
    """
    sql = "select * from coordinate_operation_view where deprecated=0 and (source_crs_auth_name='NOAA' or target_crs_auth_name='NOAA')"
    df = DB().query(sql, dataframe=True)
    failures = {"source_crs": [], "target_crs": [], "error_msg": []}
    failed_transforms = pd.DataFrame.from_dict(failures)
    for _, row in df.iterrows():
        source_crs = f"{row['source_crs_auth_name']}:{row['source_crs_code']}"
        target_crs = f"{row['target_crs_auth_name']}:{row['target_crs_code']}"
        tg = TransformerGroup(crs_from=pp.CRS(source_crs),
                              crs_to=pp.CRS(target_crs),
                              allow_ballpark=False
                              )
        if len(tg.transformers) == 0:
            error_msg = ""
            try:
                pp.Transformer.from_crs(crs_from=pp.CRS(source_crs),
                                        crs_to=pp.CRS(target_crs),
                                        allow_ballpark=False,
                                        only_best=True
                                        )

            except Exception as e:
                error_msg = str(e)
            failures["source_crs"].append(source_crs)
            failures["target_crs"].append(target_crs)
            failures["error_msg"].append(error_msg)
    failed_transforms = pd.DataFrame.from_dict(failures)
    failed_transforms.to_csv("failed_transformers.csv", index=False)
    print(f"{len(failed_transforms)}/{len(df)} NOAA transformations failed to instantiate.")



test_noaa_transforms()


import os
import pathlib
from osgeo import gdal
import glob
from vyperdatum.transformer import Transformer
from vyperdatum.utils.raster_utils import raster_metadata, update_raster_wkt
from vyperdatum.utils.vdatum_rest_utils import vdatum_cross_validate
import pyproj as pp
from pyproj import CRS, database
import math


def get_corresponding_utm_epsg(state_plane_code: str) -> str:
    """
    Given a State Plane CRS in 'AUTH:CODE' format (e.g., 'EPSG:2286'),
    returns the best matching UTM EPSG code based on extent and datum.
    Includes logging and fallback to common matches like EPSG:26910.
    """
    crs = CRS.from_user_input(state_plane_code)

    # Determine central point
    area = crs.area_of_use
    lon_center = (area.west + area.east) / 2
    lat_center = (area.south + area.north) / 2

    utm_zone = math.floor((lon_center + 180) / 6) + 1
    hemisphere = 'north' if lat_center >= 0 else 'south'

    print(f"Input CRS: {state_plane_code}")
    print(f"Estimated UTM zone: {utm_zone}, hemisphere: {hemisphere}")
    print(f"Datum: {crs.datum.name}")

    sp_datum = crs.datum.name.lower()

    candidates = []

    for info in database.query_crs_info(auth_name="EPSG"):
        name_lc = info.name.lower()
        if "utm zone" in name_lc and f"zone {utm_zone}" in name_lc and hemisphere in name_lc:
            try:
                utm_crs = CRS.from_authority(info.auth_name, info.code)
                utm_datum = utm_crs.datum.name.lower()

                print(f"Checking: {info.auth_name}:{info.code} | {utm_crs.name} | Datum: {utm_crs.datum.name}")

                if sp_datum in utm_datum or utm_datum in sp_datum:
                    print(f"Match found: {info.auth_name}:{info.code}")
                    return f"{info.auth_name}:{info.code}"
                else:
                    candidates.append((info.auth_name, info.code, utm_crs.datum.name))
            except Exception as e:
                continue

    print("No perfect datum match found.")
    print("Closest candidates:")
    for auth, code, datum in candidates:
        print(f"  - {auth}:{code} (Datum: {datum})")

    # Manual fallback: NAD83 / UTM zone 10N
    if utm_zone == 10 and hemisphere == 'north':
        print("Falling back to EPSG:26910 (NAD83 / UTM zone 10N)")
        return "EPSG:26910"

    return f"No matching UTM CRS found for {state_plane_code} (zone {utm_zone}, {hemisphere})"




def to_utm10(target_crs: str, input_tif: str, output_tif: str) -> None:
    src_ds = gdal.Open(input_tif)
    if src_ds is None:
        raise RuntimeError(f"Failed to open input file: {input_tif}")
    gdal.Warp(
        destNameOrDestDS=output_tif,
        srcDSOrSrcDSTab=src_ds,
        dstSRS=target_crs,
        format='GTiff',
        resampleAlg='bilinear',  # or nearest/cubic
        creationOptions=['COMPRESS=DEFLATE', 'TILED=YES']
    )
    return


if __name__ == "__main__":
    parent_dir = r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\CRD\Original\**\*.tif"
    files = glob.glob(parent_dir, recursive=True)[:]
    files = [r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\CRD\Original\or2014_usace_ncmp_or_dem_J1233315\or2014_usace_ncmp_or_dem_J1233315tR0_C0.tif"]
    for i, input_file in enumerate(files):
        print(f"{i+1}/{len(files)}: {input_file}")
        meta = raster_metadata(input_file)
        crs_from = f"{meta['h_authcode']}+EPSG:5703"
        crs_from = f"EPSG:6339+EPSG:5703"
        crs_to = f"EPSG:6339+NOAA:101"
        tf = Transformer(crs_from=crs_from,
                         crs_to=crs_to
                         )
        output_file = input_file.replace("Original", "Manual")
        tf.transform_raster(input_file=input_file,
                            output_file=output_file,
                            overview=False,
                            pre_post_checks=True,
                            vdatum_check=False
                            )
        print(f'\n{"*"*50} {i+1}/{len(files)} Completed {"*"*50}\n')



from vyperdatum.transformer import Transformer
from glob import glob
import numpy as np
from osgeo import gdal
import os
import tempfile
import shutil

gdal.UseExceptions()


crs_from = "EPSG:6347+EPSG:5703"
crs_to = "EPSG:6347+NOAA:98"

input_files = glob(r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\PBE\edge\Original\**\*.tif", recursive=True)
for i, input_file in enumerate(input_files):
    print(f"Processing ({i}/{len(input_files)}): {input_file}")
    output_file = input_file.replace("Original", "Manual")
    tf = Transformer(crs_from=crs_from,
                    crs_to=crs_to,
                    )
    tf.transform(input_file=input_file,
                output_file=output_file,
                pre_post_checks=True,
                vdatum_check=False
                )



import os
import vyperdatum
from enum import Enum
import pyproj as pp
from pathlib import Path


class RootEnum(Enum):
    ...


class ASSETS(Enum):
    DIR = os.path.join(os.path.split(vyperdatum.__file__)[0], "assets")


class PROJDB(RootEnum):
    """
    Proj db attributes.

    Attributes
    ----------
    FILE_NAME
    """
    # DIR = pp.datadir.get_data_dir()
    DIR = os.environ.get("VYPER_GRIDS", None)
    if DIR is None:
        raise ValueError("VYPER_GRIDS environment variable is not set.")
    if not Path(DIR).is_dir():
        raise NotADirectoryError(f"The directory specified by VYPER_GRIDS does not exist: {DIR}")
    FILE_NAME = "proj.db"
    _projdb_path = Path(DIR) / FILE_NAME
    if not _projdb_path.is_file():
        raise FileNotFoundError(f"The proj.db file does not exist in the directory specified by VYPER_GRIDS: {DIR}")
    VIEW_CRS = "crs_view"
    TABLE_VERTICAL_CRS = "vertical_crs"
    TABLE_GRID_TRANS = "grid_transformation"
    TABLE_OTHER_TRANS = "other_transformation"
    TABLE_CONCAT_OPS = "concatenated_operation"


class VDATUM(RootEnum):
    DIR = os.path.join(ASSETS.DIR.value, "vdatum")
    # from https://vdatum.noaa.gov/docs/services.html  (accessed on July 16, 2024)
    H_FRAMES = ["NAD27", "NAD83_1986", "NAD83_2011", "NAD83_NSRS2007", "NAD83_MARP00",
                "NAD83_PACP00", "WGS84_G1674", "ITRF2014", "IGS14", "ITRF2008", "IGS08",
                "ITRF2005", "IGS2005", "WGS84_G1150", "ITRF2000", "IGS00", "IGb00",
                "ITRF96", "WGS84_G873", "ITRF94", "ITRF93", "ITRF92", "SIOMIT92",
                "WGS84_G730", "ITRF91", "ITRF90", "ITRF89", "ITRF88", "WGS84_TRANSIT",
                "WGS84_G1762", "WGS84_G2139"]
    V_FRAMES = ["NAVD88", "NGVD29", "ASVD02", "W0_USGG2012", "GUVD04", "NMVD03", "PRVD02",
                "VIVD09", "CRD", "EGM2008", "EGM1996", "EGM1984", "XGEOID16B", "XGEOID17B",
                "XGEOID18B", "XGEOID19B", "XGEOID20B", "IGLD85", "LWD_IGLD85", "OHWM_IGLD85",
                "CRD", "LMSL", "MLLW", "MLW", "MTL", "DTL", "MHW", "MHHW", "LWD", "NAD27",
                "NAD83_1986", "NAD83_2011", "NAD83_NSRS2007", "NAD83_MARP00", "NAD83_PACP00",
                "WGS84_G1674", "ITRF2014", "IGS14", "ITRF2008", "IGS08", "ITRF2005", "IGS2005",
                "WGS84_G1150", "ITRF2000", "IGS00", "IGb00", "ITRF96", "WGS84_G873", "ITRF94",
                "ITRF93", "ITRF92", "SIOMIT92", "WGS84_G730", "ITRF91", "ITRF90", "ITRF89",
                "ITRF88", "WGS84_TRANSIT", "WGS84_G1762", "WGS84_G2139"]


class VRBAG(RootEnum):
    NDV_REF = 1000000
    NO_REF_INDEX = 0xffffffff

class DATUM_DOI(RootEnum):
    REGIONAL = {"url": "https://zenodo.org/records/14201516/files/regional.zip?download=1",
                "dir_name": "regional"
                }
    NWLD = {"url": "https://zenodo.org/records/15184045/files/proj.zip?download=1",
            "dir_name": "proj"
            }



import os
from typing import Optional, List, Union
from pathlib import Path
import numpy as np
import pandas as pd
import geopandas as gpd
from shapely.geometry import Point
from vyperdatum.transformer import Transformer


class XYZ:
    def __init__(self,
                 input_file: str,
                 skiprows: Optional[int] = None,
                 col_names: Optional[List[str]] = None,
                 negate_z: bool = False,
                 unit_conversion: float = 1.0
                 ):
        self.input_file = input_file
        self.skiprows = self._detect_data_start() if skiprows is None else skiprows
        self.col_names = col_names
        self.negate_z = negate_z
        self.unit_conversion = unit_conversion
        self.df = self._parse()

    def _detect_data_start(self) -> Optional[int]:
        with open(self.input_file, "r") as f:
            for i, line in enumerate(f):
                parts = line.strip().split(",")
                if len(parts) >= 3:
                    try:
                        float(parts[0])
                        float(parts[1])
                        float(parts[2])
                        return i
                    except ValueError:
                        continue
        raise ValueError("No valid data lines found in file.")

    def _parse(self) -> pd.DataFrame:
        """
        Reads the .XYZ file into a pandas DataFrame.
        If self.col_names is not set, assumes the first three columns are x, y, z.

        Returns:
            pd.DataFrame: the parsed data
        """
        df = pd.read_csv(self.input_file, sep=",", skiprows=self.skiprows)
        num_cols = df.shape[1]
        if self.col_names:
            base_names = self.col_names
        else:
            base_names = ["x", "y", "z"]
        if num_cols > len(base_names):
            column_names = base_names + [f"col{i}" for i in range(len(base_names)+1, num_cols + 1)]
        else:
            column_names = base_names[:num_cols]
        df.columns = column_names
        if self.negate_z:
            df["z"] = -df["z"]
        df["z"] *= self.unit_conversion
        return df

    def transform(self,
                  crs_from: str,
                  crs_to: str,
                  steps: Optional[List[dict]] = None
                  ) -> pd.DataFrame:
        """
        Transform the coordinates from one CRS to another using a Transformer.

        Parameters:
        -----------
            crs_from (str): The source coordinate reference system.
            crs_to (str): The target coordinate reference system.
        """
        tf = Transformer(crs_from=crs_from, crs_to=crs_to, steps=steps)
        x, y, z = self.df["x"].values, self.df["y"].values, self.df["z"].values
        success, xt, yt, zt = tf.transform_points(x, y, z,
                                                  always_xy=True,
                                                  allow_ballpark=False,
                                                  only_best=True,
                                                  vdatum_check=False)
        if not success:
            raise ValueError("Transformation failed.")
        self.df["x_t"], self.df["y_t"], self.df["z_t"] = xt, yt, zt
        if self.negate_z:
            self.df["z"] = -self.df["z"]
            self.df["z_t"] = -self.df["z_t"]
        self.df["Uncertainty"] = 1 + 0.02 * self.df["z_t"].abs()
        return self.df

    def to_gpkg(self,
                crs: str,
                output_file: str) -> None:
        tdf = pd.DataFrame({"x": self.df["x_t"].values,
                            "y": self.df["y_t"].values,
                            "Elevation": self.df["z_t"].values,
                            })
        tdf["geometry"] = tdf.apply(lambda row: Point(row["x"], row["y"], row["Elevation"]), axis=1)
        tdf["Uncertainty"] = self.df["Uncertainty"].values
        gdf = gpd.GeoDataFrame(tdf, geometry="geometry", crs=crs)
        gdf.to_file(output_file, driver="GPKG")
        return


# from vyperdatum.pipeline import Pipeline
# import sys
# # print(Pipeline(crs_from="ESRI:103060", crs_to="EPSG:6318").graph_steps())
# print(Pipeline(crs_from="EPSG:6783", crs_to="EPSG:6344").graph_steps())
# sys.exit()



input_file = r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\point\MLG\Original\AR_01_BAR_20240117_PR\AR_01_BAR_20240117_PR.XYZ"
crs_from = "EPSG:3452+NOAA:1502" # MLG depth not usable for now, use NOAA:1502 USACE height
crs_to = "EPSG:6344+NOAA:100"  # MLLW depth (due to current db issues, I use NOAA:101 and negate to get depth)
negate_z = True  # MLG depth is negative, NOAA:98 is positive


steps = [{"crs_from": "EPSG:3452", "crs_to": "EPSG:6318", "v_shift": False},
         {"crs_from": "EPSG:6318+NOAA:1502", "crs_to": "EPSG:6318+NOAA:101", "v_shift": True},
         {"crs_from": "EPSG:6318", "crs_to": "EPSG:6344", "v_shift": False}
         ]

xyz = XYZ(input_file=input_file,
          negate_z=negate_z,
          unit_conversion=0.3048006096
        #   skiprows=15,
        #   col_names=["xi", "yi", "zi"]
          )
df = xyz.transform(crs_from=crs_from, crs_to=crs_to,
                   steps=steps
                   )
output_file = input_file.replace("Original", "Manual").replace(".XYZ", ".gpkg")
os.makedirs(os.path.dirname(output_file), exist_ok=True)
xyz.to_gpkg(crs=crs_to, output_file=output_file)

print(df.head())





import os
from typing import Optional
import logging
import subprocess
import json
import pdal
from vyperdatum.drivers.base import Driver
from vyperdatum.utils.crs_utils import auth_code


logger = logging.getLogger("root_logger")


class PDAL(Driver):
    def __init__(self, input_file: str, output_file: str, invalid_error: bool = True) -> None:
        """
        Load a pdal-supported file or a remote point cloud format
        such as Entwine Pointe Tile (EPT).

        Parameters
        ----------
        input_file: str
            Full file path or address to a remote point cloud format
            such as Entwine Pointe Tile (EPT).
        output_file: str
            Full file path for the output file.
        invalid_error: bool, default True
            If True, throws an error when the input file has an unexpected format.

        Raises
        --------
        ValueError:
            If the input file is not recognized.

        Returns
        -----------
        None
        """
        super().__init__()
        self.input_file = input_file
        self.output_file = output_file
        if not self.input_file.lower().startswith("http") and not os.path.isfile(self.input_file):
            raise FileNotFoundError(f"The input file not found at {self.input_file}.")
        _, err = self.info(fname=self.input_file, args=["--schema"])
        self._is_valid = ~bool(err)
        if err and invalid_error:
            msg = f"PDAL doesn't recognize or cannot access the input: {self.input_file}"
            logger.exception(msg)
            raise ValueError(msg)
        return

    def info(self,
             fname: str,
             args: Optional[list[str]] = None) -> tuple[Optional[dict], Optional[str]]:
        """
        Spawn a new process to run `pdal info` and capture its output.

        Parameters
        -----------
        fname: str
            Path to the input file.
        args: Optional[list[str]]
            Optional pdal info arguments.

        Returns
        --------
        stdout: Optional[dict], std_err: Optional[str]
            standard output and error.
        """
        try:
            sout, serr = dict({}), None
            resp = subprocess.run(["pdal", "info", *args, fname],
                                  stderr=subprocess.PIPE,
                                  stdout=subprocess.PIPE
                                  )
            sout = json.loads(resp.stdout.decode()) if resp.stdout else None
            serr = resp.stderr.decode() if resp.stderr else None
        except Exception as e:
            logger.exception(str(e))
            sout, serr = dict({}), None
        return sout, serr

    def exec_pipeline(self, pipe_string: str):
        """
        Execute a general pdal pipeline.

        Parameters
        -----------
        pipe_string: str
            PDAL's pipeline json object in string format.
        """
        try:
            pipe_string = os.linesep.join([s for s in pipe_string.splitlines() if s])
            p = pdal.Pipeline(pipe_string)
            count = p.execute()
            arrays = p.arrays
            metadata = p.metadata
            log = p.log
        except Exception as e:
            logger.exception(str(e))
        return

    def fetch_entwine(self,
                      bounds: Optional[list[float]],
                      resolution: Optional[int],
                      output_format: str = "writers.las"
                      ) -> bool:
        """
        Download a subset of Etwine resource defined by `bounds: [x_min, x_max, y_min, y_max]`.
        The subset data is converted into the `output_format` which is a
        PDAL type.

        Parameters
        -----------
        bounds: Optional[list[float]]
            Bounding box that delimits the remote Etwine resource to be downloaded.
            It should be a list of float values in this order: [x_min, x_max, y_min, y_max].
        output_format: str, default 'writers.las'
            PDAL writers type; defaults to 'writers.las'.
            See more https://pdal.io/en/2.8.1/stages/writers.html.

        Returns
        --------
        bool
            True if download succeed, otherwise False.
        """
        bounds = f',"bounds": "([{bounds[0]}, {bounds[1]}], [{bounds[2]}, {bounds[3]}])"' if bounds else ''
        resolution = f',"resolution": {resolution}' if resolution else ''
        try:
            succeed = False
            out_file = self.output_file.replace("\\", "/")
            pip_string = f'''
                        [
                        {{
                        "type":"readers.ept",
                        "filename":"{self.input_file}"
                        {bounds}
                        {resolution}
                        }},
                        {{
                        "type":"{output_format}",
                        "filename":"{out_file}"
                        }}
                        ]
                        '''
            self.exec_pipeline(pipe_string=pip_string)
            succeed = os.path.isfile(self.output_file)
        except Exception as e:
            succeed = False
            logger.error(f"Error in fetch_entwine: {e}")
        return succeed

    def wkt(self, fname: str) -> str:
        """
        Return the WKT associated with the input file `fname`.

        Parameters
        -----------
        fname: str
            Path to the input file.

        Raises
        -----------
        ValueError
            When `pdal info --metadata <fname>` fails.

        Returns
        -----------
        str
        """
        resp, err = self.info(fname=fname, args=["--metadata"])
        if bool(err):
            raise ValueError(f"Error in getting metadata: {err}")
        # expected keys in metadata.srs: ['compoundwkt', 'horizontal', 'isgeocentric', 'isgeographic', 'json', 'prettycompoundwkt', 'prettywkt', 'proj4', 'units', 'vertical', 'wkt']
        return resp["metadata"]["srs"]["compoundwkt"]

    @property
    def is_valid(self):
        return self._is_valid

    def transform(self, transformer_instance, vdatum_check: bool) -> None:
        """
        Create a PDAL pipeline to apply CRS transformation on the input data
        according to the `transformer_instance`.

        Parameters
        -----------
        transformer_instance: vyperdatum.transformer.Transform
            Instance of the transformer class.

        Returns
        -----------
        None
        """
        in_file = self.input_file.replace("\\", "/")
        out_file = self.output_file.replace("\\", "/")
        pip_string = f'''
                      [
                      "{in_file}",
                      {{
                      "type":"filters.reprojection",
                      "out_srs":"{auth_code(transformer_instance.crs_to)}"
                      }},
                      "{out_file}"
                      ]
                      '''
        self.exec_pipeline(pipe_string=pip_string)
        return


if __name__ == "__main__":
    from vyperdatum.transformer import Transformer
    import pyproj as pp
    # fname = r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\raster\BlueTopo_BC25L26L_20230919.tiff"
    # fname = "https://na-c.entwine.io/dublin/ept.json"
    fname = r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\point\laz\ma2021_cent_east_Job1082403.laz"
    oname = r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\point\laz\_t_ma2021_cent_east_Job1082403.laz"
    p = PDAL(input_file=fname, output_file=oname)
    crs_from = "EPSG:6348+EPSG:5703"
    crs_to = "EPSG:6348+NOAA:5320"
    steps = ["EPSG:6348+EPSG:5703", "EPSG:6318+EPSG:5703", "EPSG:6318+NOAA:5320", "EPSG:6348+NOAA:5320"]
    tf = Transformer(crs_from=crs_from, crs_to=crs_to, steps=steps)
    p.transform(tf)
    wkt = p.wkt(fname=fname)
    print(wkt)
    # print(pp.CRS(wkt).to_authority())


    # pdl = PDAL(input_file="https://na-c.entwine.io/dublin/ept.json",
    #            output_file=r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\pdal\entwine.las")
    # succeed = pdl.fetch_entwine(bounds=None, resolution=20, output_format="writers.las")
    # assert succeed, "fetch_entwine failed!"



import os
import glob
import logging
from typing import Optional, Union
import pandas as pd
from osgeo import gdal, ogr
from vyperdatum.db import DB


logger = logging.getLogger("root_logger")
gdal.UseExceptions()


def get_grid_list(vdatum_directory: str):
    """
    Return all gtx files in the vdatum directory.

    Parameters
    ----------
    vdatum_directory
        absolute folder path to the vdatum directory

    Returns
    -------
    dict
        dictionary of {grid name: grid path, ...}
    list
        list of vdatum regions
    """
    grid_formats = [".tif", ".tiff", ".gtx"]
    grid_list = []
    for gfmt in grid_formats:
        search_path = os.path.join(vdatum_directory, "*/*{}".format(gfmt))
        grid_list += glob.glob(search_path)
    if len(grid_list) == 0:
        logger.error(f"No grid files found in the provided VDatum directory: {vdatum_directory}")
    grids = {}
    regions = []
    for grd in grid_list:
        grd_path, grd_file = os.path.split(grd)
        grd_path, grd_folder = os.path.split(grd_path)
        gtx_name = "/".join([grd_folder, grd_file])
        gtx_subpath = os.path.join(grd_folder, grd_file)
        grids[gtx_name] = gtx_subpath
        regions.append(grd_folder)
    regions = list(set(regions))
    return grids, regions


def get_region_polygons(datums_directory: str, extension: str = "kml") -> dict:
    """"
    Search the datums directory to find all geometry files.
    All datums are assumed to reside in a subfolder.

    Parameters
    ----------
    datums_directory : str
        absolute folder path to the vdatum directory

    extension : str
        the geometry file extension to search for

    Returns
    -------
    dict
        dictionary of {kml name: kml path, ...}
    """

    search_path = os.path.join(datums_directory, f"*/*.{extension}")
    geom_list = glob.glob(search_path)
    if len(geom_list) == 0:
        logger.error(f"No {extension} files found in the provided directory: {datums_directory}")
    geom = {}
    for filename in geom_list:
        geom_path, _ = os.path.split(filename)
        _, geom_name = os.path.split(geom_path)
        geom[geom_name] = filename
    return geom


def overlapping_regions(datums_directory: str,
                        lon_min: float,
                        lat_min: float,
                        lon_max: float,
                        lat_max: float
                        ) -> list[str]:
    """
    Return the region names that intersect with the provided bound.
    The input coordinate reference system is expected to be
    NAD83(2011) geographic.

    Parameters
    ----------
    lon_min
        the minimum longitude of the area of interest
    lat_min
        the minimum latitude of the area of interest
    lon_max
        the maximum longitude of the area of interest
    lat_max
        the maximum latitude of the area of interest

    Returns
    ----------
    list[str]
    """
    assert lon_min < lon_max
    assert lat_min < lat_max

    # build corners from the provided bounds
    ul = (lon_min, lat_max)
    ur = (lon_max, lat_max)
    lr = (lon_max, lat_min)
    ll = (lon_min, lat_min)

    # build polygon from corners
    ring = ogr.Geometry(ogr.wkbLinearRing)
    ring.AddPoint(ul[0], ul[1])
    ring.AddPoint(ur[0], ur[1])
    ring.AddPoint(lr[0], lr[1])
    ring.AddPoint(ll[0], ll[1])
    ring.AddPoint(ul[0], ul[1])
    data_geometry = ogr.Geometry(ogr.wkbPolygon)
    data_geometry.AddGeometry(ring)

    # see if the regions intersect with the provided geometries
    intersecting_regions = []
    polygon_files = get_region_polygons(datums_directory)
    for region in polygon_files:
        vector = ogr.Open(polygon_files[region])
        layer_count = vector.GetLayerCount()
        # found = False
        for m in range(layer_count):
            layer = vector.GetLayerByIndex(m)
            feature_count = layer.GetFeatureCount()
            for n in range(feature_count):
                feature = layer.GetNextFeature()
                try:
                    feature_name = feature.GetField(0)
                except AttributeError:
                    logger.warning("WARNING: Unable to read feature name from feature"
                                   f"in layer in {polygon_files[region]}"
                                   )
                    continue
                if isinstance(feature_name, str):
                    if feature_name[:15] == "valid-transform":
                        valid_vdatum_poly = feature.GetGeometryRef()
                        if data_geometry.Intersect(valid_vdatum_poly):
                            intersecting_regions.append(region)
                            # found = True
                feature = None
            layer = None
        # if not found and region in self.datum_data.extended_region:
        #     feature = vector.GetLayerByIndex(0).GetFeature(0)
        #     if data_geometry.Intersect(feature.GetGeometryRef()):
        #         intersecting_regions.append(region)
        vector = None
    return intersecting_regions


def overlapping_extents(lon_min: float,
                        lat_min: float,
                        lon_max: float,
                        lat_max: float
                        ) -> Union[Optional[list], Optional[pd.DataFrame]]:
    """
    Return database predefined extents that intersect with an area of interest.
    The results are sorted by coverage ratios.

    Parameters
    ----------
    lon_min
        the minimum longitude of the area of interest
    lat_min
        the minimum latitude of the area of interest
    lon_max
        the maximum longitude of the area of interest
    lat_max
        the maximum latitude of the area of interest
    """

    sql = f"""
    with data_extent AS (
        SELECT {lon_min} as min_lon,
               {lon_max} as max_lon,
               {lat_min} as min_lat,
               {lat_max} as max_lat
        )
    select *,

    -- ratio of data covered by a database extent:
    -- intersecting area of data and database extent divided by data area
    (
    (
    (min(data_extent.max_lon, east_lon) - max(data_extent.min_lon, west_lon))
    *
    (min(data_extent.max_lat, north_lat) - max(data_extent.min_lat, south_lat))
    ) / ((data_extent.max_lon-data_extent.min_lon) * (data_extent.max_lat-data_extent.min_lat)) 
    )
    as data_coverage_ratio,


    -- ratio of a database extent covered by data:
    -- intersecting area of data and database extent divided by the database extent area
    (
    (
    (min(data_extent.max_lon, east_lon) - max(data_extent.min_lon, west_lon))
    *
    (min(data_extent.max_lat, north_lat) - max(data_extent.min_lat, south_lat))
    ) / ((east_lon-west_lon) * (north_lat-south_lat)) 
    )
    as extent_coverage_ratio

    from extent, data_extent
    where
    east_lon >= data_extent.min_lon
    and west_lon <= data_extent.max_lon
    and north_lat >= data_extent.min_lat
    and south_lat <= data_extent.max_lat
    and deprecated = 0
    order by data_coverage_ratio desc,
             extent_coverage_ratio desc,
             (abs(east_lon - west_lon) * abs(north_lat - south_lat))
    """

    return DB().query(sql, dataframe=True)



import pytest
import pyproj as pp
from vyperdatum.drivers.pdal_based import PDAL
from vyperdatum.transformer import Transformer
from vyperdatum.utils.crs_utils import auth_code


def test_fetch_entwine():
    pdl = PDAL(input_file="https://na-c.entwine.io/dublin/ept.json",
               output_file=r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\pdal\entwine.las")
    succeed = pdl.fetch_entwine(bounds=None, resolution=20, output_format="writers.las")
    assert succeed, "fetch_entwine failed!"
    return


def test_transform():
    fname = r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\point\laz\ma2021_cent_east_Job1082403.laz"
    oname = r"C:\Users\mohammad.ashkezari\Documents\projects\vyperdatum\untrack\data\point\laz\_t_ma2021_cent_east_Job1082403.laz"
    p = PDAL(input_file=fname, output_file=oname)
    crs_from = "EPSG:6348+EPSG:5703"
    crs_to = "EPSG:6348+NOAA:98"
    tf = Transformer(crs_from=crs_from, crs_to=crs_to)
    p.transform(tf)
    wkt = p.wkt(fname=oname)
    assert auth_code(pp.CRS(wkt)) == "EPSG:6348+NOAA:98", ("Transformed CRS"
                                                             f" code is {auth_code(pp.CRS(wkt))}"
                                                             " but expected 'EPSG:6348+NOAA:98'")
    return
